/**********************************************************
 **
 ** bsc_expr.c : BSC expression parsing
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include "ptype.h"
#include "pfile.h"
#include "expr.h"
#include "bsc_tokn.h"
#include "bsc_expr.h"

/*
 * NAME
 *   bsc_subscript_parse
 *
 * DESCRIPTION
 *   get the subscript for an array variable
 *
 * PARAMETERS
 *   pf  : pfile
 *   val : value
 *
 * RETURN
 *   EINVAL : syntax error
 *   
 * NOTES
 *   the pfile token should be '(' on entry
 */
result_t bsc_subscript_parse(bsc_file_t *bsc, value_t *val)
{
  result_t rc;

  if (!bsc_token_is(bsc, bsc_token_current, pfile_log_err, "(")) {
    rc = result_syntax;
  } else {
    value_t *subscript;

    bsc_token_get(bsc , bsc_token_next);
    rc = bsc_expr_parse(bsc, 0, &subscript);
    if (result_ok == rc) {
      variable_sz_t basesz;

      basesz = variable_flag_test(value_variable_get(val), VARIABLE_FLAG_CONST)
          ? 1
          : variable_sz_get(value_variable_get(val));
      if (value_flag_test(subscript, VARIABLE_FLAG_CONST)) {
        /* this is a constant subscript, adjust the baseofs */
        variable_const_t n;

        n = value_const_get(subscript);
        if (n >= value_ct_get(val)) {
          pfile_log(bsc_file_pf_get(bsc), pfile_log_err, 
            BSC_MSG_SUBSCRIPT_BAD);
          rc = result_range;
        } else {
          value_ct_ofs_set(val, n);
        }
      } else {
        /* if subscript is not a temporary, or variable sz > 1,
           we'll need a temporary */
        if (basesz > 1) {
          value_t *tmp;
          value_t *tmp_sz;

          tmp    = 0;
          tmp_sz = 0;
          rc = pfile_constant_get(bsc_file_pf_get(bsc), basesz, &tmp_sz);
          if (result_ok == rc) {
            rc = pfile_value_temp_get(bsc_file_pf_get(bsc), 2, 0, &tmp);
          }
          if (result_ok == rc) {
            pfile_cmd_op_add(bsc_file_pf_get(bsc), 0, operator_assign, &tmp, 
                subscript, 0);
            pfile_cmd_op_add(bsc_file_pf_get(bsc), 0, operator_mul, &tmp, 
                tmp, tmp_sz);
          }
          value_release(subscript);
          value_release(tmp_sz);
          subscript = tmp;
        }
        if (result_ok == rc) {
          value_subscript_set(val, value_variable_get(subscript));
        }
      }
      value_release(subscript);
      if (!bsc_token_is(bsc, bsc_token_current, pfile_log_err, 
        ")")) {
          rc = result_syntax;
      } else {
        bsc_token_get(bsc, bsc_token_next);
      }
    }
  }
  return rc;
}

/*
 * NAME
 *   value_get
 *
 * DESCRIPTION
 *   get an expression's value
 *
 * PARAMETERS
 *   pf    : return from pfile_open()
 *   flags : flags passed into expr_parse()
 *   var   : [out] holds result
 *
 * RETURN
 *
 * NOTES
 */
static result_t value_get(bsc_file_t *bsc, flag_t flags, value_t **dst)
{
  result_t    rc;
  value_t    *var;
  boolean_t   skipnext;
  const char *ptr;
  operator_t  op;

  /* valid values: 
       constant ('%' or '$' or decimal #)
       variable name (alpha + alpha|digit|'_')
       '(' -- subfunction
       '!' -- logical NOT
       '-' -- negation (two's complement)
       '~' -- complement (one's complement) */

  skipnext = boolean_false;
  ptr = bsc_token_get(bsc, bsc_token_current);
  op = operator_null;
  switch (*ptr) {
    case '!':
      op = ('!' == ptr[1]) ? operator_logical : operator_notl;
      /* fall through */
    case '-':
      if (operator_null == op) {
        op = operator_neg;
      }
      /* fall through */
    case '~':
      if (operator_null == op) {
        op = operator_cmpb;
      }
      bsc_token_get(bsc, bsc_token_next); /* this one's done */
      rc = value_get(bsc, flags, &var);   /* var is the next value */
      /* create the expression:
       *   _t# = var op
       *   and return _t#
       */
      if (result_ok == rc) {
        value_t *tmp;

        tmp = 0;
        pfile_cmd_op_add(bsc_file_pf_get(bsc), 0, op, &tmp, var, 0);
        value_release(var);
        var = tmp;
      }
      skipnext = 1;
      break;
    case '(':
      bsc_token_get(bsc, bsc_token_next);
      rc = bsc_expr_parse(bsc, flags, &var);
      /* prepend the existing result to the new result */
      if (!bsc_token_is(bsc, bsc_token_current, pfile_log_err, ")")
          && (result_ok == rc)) {
        rc = result_syntax;
      }
      break;
    default:
      rc = bsc_token_to_constant(bsc, pfile_log_none, &var);
      if (result_invalid == rc) {
        rc = bsc_token_to_value(bsc, pfile_log_none, &var);
      }
      if (result_ok != rc) {
        pfile_log(bsc_file_pf_get(bsc), pfile_log_err, BSC_MSG_VALUE_EXPECTED);
      } else if ((result_ok == rc) && value_is_array(var)) {
        /* need a subscript */
        bsc_token_get(bsc, bsc_token_next);
        rc = bsc_subscript_parse(bsc, var);
        if (result_ok != rc) {
          value_release(var);
        } 
        skipnext = boolean_true;
      }
      break;
  }
  if (result_ok == rc) {
    *dst = var;
    if (!skipnext) {
      bsc_token_get(bsc, bsc_token_next);
    }
  }
  return rc;
}

/*
 * NAME
 *   operator_get
 *
 * DESCRIPTION
 *   get an operator
 *
 * PARAMETERS
 *   expr : expression
 *   pf   : returned by pfile_open()
 *
 * RETURN
 *   0      : no error
 *   EINVAL : no operator present
 *
 * NOTES
 */
static operator_t operator_get(bsc_file_t *bsc, flag_t flags)
{
  static const struct {
    const char *token;
    operator_t     op;
  } operator_token_map[] = {
    {"+",  operator_add},
    {"-",  operator_sub},
    {"*",  operator_mul},
    {"/",  operator_div},
    {"<",  operator_lt},
    {"<=", operator_le},
    {"=",  operator_eq},
    {"<>", operator_ne},
    {"><", operator_ne},
    {">=", operator_ge},
    {">",  operator_gt},
    {"&",  operator_andb},
    {"|",  operator_orb},
    {"^",  operator_xorb},
    {"!!", operator_logical},
    {"mod", operator_mod},
    {"<<",  operator_shift_left},
    {">>",  operator_shift_right},
    {"&&",  operator_andl},
    {"||",  operator_orl},
    {"andb", operator_andb},
    {"orb",  operator_orb}
  };
  size_t ii;
  operator_t op;

  for (ii = 0; 
       (ii < COUNT(operator_token_map))
        && !bsc_token_is(bsc, bsc_token_current, pfile_log_none,
          operator_token_map[ii].token); 
        ii++)
    ;
  if (ii == COUNT(operator_token_map)) {
    op = operator_null;
  } else {
    op = operator_token_map[ii].op;
    if (flags & EXPR_FLAG_IF) {
      if (bsc_token_is(bsc, bsc_token_current, pfile_log_none, "&")) {
        op = operator_andl;
      } else if (bsc_token_is(bsc, bsc_token_current, pfile_log_none, "|")) {
        op = operator_orl;
      }
    }
    bsc_token_get(bsc, bsc_token_next);
  }
  return op;
}

/*
 * NAME
 *   expr_parse
 *
 * DESCRIPTION
 *   parse an expression
 *
 * PARAMETERS
 *   dst : [out] holds result
 *   pf  : parse file
 *
 * RETURN
 *
 * NOTES
 */
result_t bsc_expr_parse(bsc_file_t *bsc, flag_t flags, value_t **dst)
{
  operator_t  op;
  expr_t     *exprstk;
  value_t    *var;
  result_t    rc;

  exprstk = 0;
  op      = operator_null;
  do {
    rc = value_get(bsc, flags, &var);
    if (result_ok == rc) {
      op = operator_get(bsc, flags);
      /* if this is in an ``if'' statement, all binary ops become
         logical */
      if (flags & EXPR_FLAG_IF) {
        if (operator_andb == op) {
          op = operator_andl;
        } else if (operator_orb == op) {
          op = operator_orl;
        }
      }
      rc = pfile_expr_push(bsc_file_pf_get(bsc), &exprstk, var, op);
      value_release(var);
    }
  } while ((operator_null != op) && (result_ok == rc));
  if (exprstk) {
    if (result_ok == rc) {
      var = expr_val_get(exprstk);
      value_lock(var);
      *dst = var;
    }
    expr_list_free(&exprstk);
  }
  return rc;
}

