/************************************************************
 **
 ** bsc_file.c : BSC file structure definitions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <string.h>
#include "pfile.h"
#include "bsc_tokn.h"
#include "bsc_file.h"
#include "bsc_decl.h"

result_t bsc_file_alloc(bsc_file_t **dst, pfile_t *pf)
{
  bsc_file_t *bsc;
  result_t    rc;

  bsc = MALLOC(sizeof(*bsc));
  if (!bsc) {
    rc = result_memory;
  } else {
    size_t ii;

    bsc->pf       = pf;
    bsc->pass     = 0;
    bsc->str_ct   = 0;
    bsc->str_nxt  = 0;
    bsc->inc_path = 0;

    for (ii = 0; ii < COUNT(bsc->isrs); ii++) {
      bsc->isrs[ii] = 0;
    }
    bsc->isrs[bsc_isr_type_ct] = pfile_label_find(pf, pfile_log_err, 
      "_isr_cleanup");
    if (!bsc->isrs[bsc_isr_type_ct]) {
      FREE(bsc);
      rc = result_internal;
    } else {
      rc = result_ok;
      *dst = bsc;
    }
  }
  return rc;
}

void bsc_file_free(bsc_file_t *bsc)
{
  size_t ii;

  for (ii = 0; ii < COUNT(bsc->isrs); ii++) {
    if (bsc->isrs[ii]) {
      label_release(bsc->isrs[ii]);
    }
  }
  FREE(bsc);
}


pfile_t *bsc_file_pf_get(bsc_file_t *bsc)
{
  return bsc->pf;
}

result_t bsc_file_isr_get(bsc_file_t *bsc, bsc_isr_type_t isr, 
    boolean_t alloc, label_t **dst)
{
  label_t *lbl;
  result_t rc;

  lbl = bsc->isrs[isr];
  if (lbl) {
    rc = result_ok;
  } else if (!alloc) {
    rc = result_not_found;
  } else {
    const char *str;

    str = "{unknown}";
    switch (isr) {
      case bsc_isr_type_adc:   str = "_ir_adc";   break;
      case bsc_isr_type_pwm:   str = "_ir_pwm";   break;
      case bsc_isr_type_rx:    str = "_ir_rx";    break;
      case bsc_isr_type_tx:    str = "_ir_tx";    break;
      case bsc_isr_type_timer: str = "_ir_timer"; break;
      case bsc_isr_type_rb:    str = "_ir_rb";    break;
      case bsc_isr_type_user:  str = "_ir_user";  break;
      case bsc_isr_type_ct: break;
    }
    if (str) {
      rc = pfile_label_alloc(bsc_file_pf_get(bsc), str, 0, &lbl);
      if (result_ok == rc) {
        bsc->isrs[isr] = lbl;
      }
    } else {
      rc = result_internal;
    }
  }
  if (result_ok == rc) {
    if (dst) {
      label_lock(lbl);
      *dst = lbl;
    }
  }
  return rc;
}

label_t *bsc_file_isr_next_get(bsc_file_t *bsc, bsc_isr_type_t type)
{
  while ((++type < COUNT(bsc->isrs)) && !bsc->isrs[type]) 
    ;
  return (type < COUNT(bsc->isrs)) ? bsc->isrs[type] : 0;
}

void bsc_file_pass_reset(bsc_file_t *bsc)
{
  pfile_reset(bsc_file_pf_get(bsc));
  bsc->pass++;
  bsc_token_get(bsc, bsc_token_first);
  bsc->str_ct = 0;
}

bsc_pass_t  bsc_file_pass_get(bsc_file_t *bsc)
{
  return bsc->pass;
}

const char *bsc_token_ptr_get(bsc_file_t *bsc)
{
  return bsc->token;
}

void bsc_token_set(bsc_file_t *bsc, const char *token)
{
  strcpy(bsc->token, token);
}

void bsc_token_ch_append(bsc_file_t *bsc, int ch)
{
  if (EOF != ch) {
    if (bsc->token_len + 1 < sizeof(bsc->token)) {
      bsc->token[bsc->token_len++] = ch;
      bsc->token[bsc->token_len] = 0;
    }
  }
}

void bsc_token_reset(bsc_file_t *bsc)
{
  bsc->token_len = 0;
  bsc->token[0] = 0;
}

size_t bsc_file_strid_next(bsc_file_t *bsc)
{
  return bsc->str_ct++;
}

void bsc_file_log_unavail(bsc_file_t *bsc, const char *cmd)
{
  pfile_log(bsc_file_pf_get(bsc), pfile_log_err, 
      BSC_MSG_CMD_NOT_AVAIL, cmd);
}

void bsc_file_include_path_set(bsc_file_t *bsc, const char *inc_path)
{
  bsc->inc_path = inc_path;
}

const char *bsc_file_include_path_get(const bsc_file_t *bsc)
{
  return bsc->inc_path;
}

boolean_t bsc_file_flag_test(const bsc_file_t *bsc, bsc_file_flag_t flag)
{
  return (bsc->flags & (1 << flag)) != 0;
}

void bsc_file_flag_set(bsc_file_t *bsc, bsc_file_flag_t flag)
{
  bsc->flags |= (1 << flag);
}

void bsc_file_flag_clr(bsc_file_t *bsc, bsc_file_flag_t flag)
{
  bsc->flags &= ~(1 << flag);
}

