/**********************************************************
 **
 ** bsc_proc.c : BSC PROC/CALL/IR_* parsing
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include <stdlib.h>
#include <string.h>

#include "pfile.h"
#include "label.h"
#include "value.h"
#include "bsc_tokn.h"
#include "bsc_blck.h"
#include "bsc_expr.h"
#include "bsc_proc.h"

#define PROC_PARAMS_MAX 100

typedef struct procdef_ {
  size_t    param_ct; /* # of parameters                        */
  value_t **params;   /* parameter names (these are NOT locked) */
} procdef_t;

static void lblproc_free(label_t *lbl)
{
  procdef_t *data;

  data = label_data_get(lbl);
  if (data) {
    while (data->param_ct--) {
      value_release(data->params[data->param_ct]);
    }
    FREE(data);
  }
}

/*
 * NAME
 *   bsc_parse_proc
 *
 * DESCRIPTION
 *   define a procedure
 *
 * PARAMETERS
 *   pf       : pfile
 *   sub      : sub fn (== 0)
 *   param_ct : # of parameters (1)
 *   params   : parameter (label)
 *
 * RETURN
 *   none
 *
 * NOTES
 *      parse:
 *      label (need not exist)
 *      '(' [var1 [, var2...]] ')'
 *      block
 *      endproc
 */
void bsc_parse_proc(bsc_file_t *bsc, unsigned sub, size_t param_ct, 
  const param_t *params)
{
  static const char *ewords[] = { "endproc" };
  label_t *proc_end;
  label_t *lbl;
  pfile_t *pf;

  pf = bsc_file_pf_get(bsc);
  lbl = 0;
  if (1 == bsc_file_pass_get(bsc)) {
    /* all this is only needed on the first pass,
       after that, skip it! */
    result_t    rc;
    procdef_t  *data;
    boolean_t   is_new;

    rc = bsc_token_to_label(bsc, pfile_log_none, boolean_false, &lbl);
    if (result_ok == rc) {
      if (label_flag_test(lbl, LABEL_FLAG_DEFINED)) {
        pfile_log(pf, pfile_log_err, BSC_MSG_MULTI_DEF, 
          bsc_token_ptr_get(bsc));
      }
    } else if (result_not_found == rc) {
      rc = pfile_label_alloc(pf, bsc_token_ptr_get(bsc), lblproc_free, &lbl);
    }
    if (result_ok == rc) {
      is_new = !label_flag_test(lbl, LABEL_FLAG_DEFINED);
      if (is_new) {
        label_flag_set(lbl, LABEL_FLAG_DEFINED);
        data = MALLOC(sizeof(*data) + PROC_PARAMS_MAX * sizeof(*data->params));
        if (0 == data) {
          pfile_log_syserr(pf, result_memory);
        } else {
          data->param_ct = 0;
          data->params   = (void *) (data + 1);
          label_data_set(lbl, data);
        }
      } else {
        data = 0;
      }
      if (bsc_token_is(bsc, bsc_token_next, pfile_log_err, "(")) {
        size_t ii;

        for (ii = 0;
            (ii < PROC_PARAMS_MAX)
            && (!ii
              || bsc_token_is(bsc, bsc_token_current, pfile_log_none, ","));
            ii++) {
          value_t *var;

          /* skip the ',' or '(' */
          bsc_token_get(bsc, bsc_token_next);

          if (result_ok == bsc_token_to_value(bsc, 
                (ii) ? pfile_log_err : pfile_log_none, &var)) {
            bsc_token_get(bsc, bsc_token_next);
            if (value_is_const(var)) {
              pfile_log(pf, pfile_log_err, BSC_MSG_VARIABLE_EXPECTED);
              value_release(var);
              var = 0;
            } else if (data) {
              /* this is not optimal as a parameter may never be
               * assigned or used, but I've no choice */
              value_flag_set(var, VARIABLE_FLAG_USED);
              value_flag_set(var, VARIABLE_FLAG_ASSIGNED);
              data->params[ii] = var;
              value_assign_ct_bump(var, ctr_bump_incr);
            } else {
              value_release(var);
            }
          } else /* if (0 == ii) */ {
            break; /* null parameter list, bail! */
          }
          if (!var) {
            break;
          }
        }
        if (data) {
          data->param_ct = ii;
        }
        if (bsc_token_is(bsc, bsc_token_current, pfile_log_err, ")")) {
          bsc_token_get(bsc, bsc_token_next);
        }
      }
    }
  } else {
    bsc_token_to_label(bsc, pfile_log_err, boolean_false, &lbl);
    bsc_statement_skip(bsc);
    if (!label_flag_test(lbl, LABEL_FLAG_USED)) {
      pfile_log(pf, pfile_log_warn, BSC_MSG_NEVER_USED,
          label_name_get(lbl));
    }
  }
  /* process the block */
  if (result_ok != pfile_label_alloc(pf, 0, 0, &proc_end)) {
    proc_end = 0;
  }
  if (proc_end) {
    pfile_cmd_branch_add(pf, 0, cmd_branchtype_goto, 
      cmd_branchcond_none, proc_end, 0);
  }
  if (lbl) {
    pfile_cmd_label_add(pf, lbl);
  }
  bsc_block_process(bsc, 1, COUNT(ewords), ewords, 0);
  pfile_cmd_branch_add(pf, 0, cmd_branchtype_return,
    cmd_branchcond_none, 0, 0);
  if (proc_end) {
    pfile_cmd_label_add(pf, proc_end);
    label_release(proc_end);
  }
  label_release(lbl);
}

/*
 * NAME
 *   bsc_parse_call
 *
 * DESCRIPTION
 *   call a procedure
 *
 * PARAMETERS
 *   pf       : pfile
 :   sub      : sub fn ( == 0)
 *   param_ct : # of parameters (0)
 *   params   : parameters (0)
 *
 * RETURN
 *   none
 *
 * NOTES
 *   format: call proc( [expr1 [,expr2...]] )
 *   code:
 *     proc.param[0] = expr1
 *     proc.param[1] = expr2
 *     ...
 *     call procname
 *     for each parameter that was a variable, set it back
 */
static int is_temporary(const value_t *var)
{
  return !value_name_get(var)
    || (0 == memcmp(value_name_get(var), "_t", 2));
}

void bsc_parse_call(bsc_file_t *bsc, unsigned sub, size_t param_ct, 
  const param_t *params)
{
  pfile_t    *pf;
  label_t    *lbl;
  procdef_t  *def;
  value_t   **parms;
  size_t      parms_ct;

  pf = bsc_file_pf_get(bsc);

  lbl = 0;
  if (1 == bsc_file_pass_get(bsc)) {
    if (result_not_found
        == bsc_token_to_label(bsc, pfile_log_none, boolean_false, &lbl)) {
      pfile_label_alloc(pf, bsc_token_ptr_get(bsc), lblproc_free, &lbl);
    }
  } else {
    if (result_ok == bsc_token_to_label(bsc, pfile_log_err, boolean_false, 
          &lbl)) {
      /* bsc_token_get(bsc, bsc_token_next); */
    }
  }
  def      = (lbl) ? label_data_get(lbl) : 0;
  parms_ct = (def) ? def->param_ct : PROC_PARAMS_MAX;
  parms    = (def && parms_ct) ? CALLOC(sizeof(*parms), parms_ct) : 0;
  if (param_ct && def && !parms) {
    pfile_log_syserr(pf, result_memory);
  } else if (!def && (1 != bsc_file_pass_get(bsc))) {
    pfile_log(pf, pfile_log_err, BSC_MSG_NEVER_DEFINED, 
      bsc_token_ptr_get(bsc));
  }
  if (bsc_token_is(bsc, bsc_token_next, pfile_log_err, "(")) {
    size_t ct;

    if (!parms_ct) {
      bsc_token_get(bsc, bsc_token_next);
    } else {
      for (ct = 0; 
           (ct < parms_ct)
           && (!ct || bsc_token_is(bsc, bsc_token_current, 
             pfile_log_none, ","));
           ct++) {
        value_t *var;

        /* skip either '(' or ',' */
        bsc_token_get(bsc, bsc_token_next);

        if (!ct && bsc_token_is(bsc, bsc_token_current, pfile_log_none, ")")) {
          break;
        }

        if (result_ok == bsc_expr_parse(bsc, 0, &var)) {
          if (parms) {
            parms[ct] = var;
            pfile_cmd_op_add(pf, 0, operator_assign, &def->params[ct], var, 
                0);
          } else {
            /* must mark this value as used */
            /* value_use_ct_bump(var, ctr_bump_incr); */
            const char *vname;

            vname = value_name_get(var);
            if (vname && !(('_' == vname[0]) && ('t' == vname[1]))) {
              value_flag_set(var, VARIABLE_FLAG_USED);
              value_flag_set(var, VARIABLE_FLAG_ASSIGNED);
            }
            value_release(var);
          }
        } else if ((0 == ct) && !parms) {
          break; /* null parameter list, bail */
        } else if (parms) {
          parms[ct] = 0;
        }
      }
    }
    if (bsc_token_is(bsc, bsc_token_current, pfile_log_err, ")")) {
      bsc_token_get(bsc, bsc_token_next);
    }
    if (lbl) {
      pfile_cmd_branch_add(pf, 0, cmd_branchtype_call, cmd_branchcond_none,
        lbl, 0);
    }
    if (parms) {
      for (ct = 0; ct < parms_ct; ct++) {
        if (parms[ct]) {
          if ((!is_temporary(parms[ct]) && !value_is_const(parms[ct]))) {
            pfile_cmd_op_add(pf, 0, operator_assign, &parms[ct],
              def->params[ct], 0);
          }
          value_release(parms[ct]);
        }
      }
    }
    if (parms) {
      FREE(parms);
    }
    label_release(lbl);
  }
}

