/**********************************************************
 **
 ** bsc_tokn.c : BSC token parsing bits
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <errno.h>

#include "pfile.h"
#include "bsc_tokn.h"


/*
 * NAME
 *   bsc_token_parse_allow
 *
 * DESCRIPTION
 *   parse the token while the character is allowed
 *
 * PARAMETERS
 *   pf    : returned by bsc_open()
 *   allow : allow function
 *   flags : determine what to do with the characters
 *
 * RETURN
 *   first character not allowed
 *
 * NOTES
 */
int bsc_token_parse_allow(
  bsc_file_t                *bsc, 
  bsc_token_parse_allow_cb_t allow, 
  flag_t                     flags)
{
  int ch;
  pfile_t *pf;

  pf = bsc_file_pf_get(bsc);

  for ( ; ; ) {
    ch = pfile_ch_get(pf);
    if ((EOF == ch) || !allow(ch)) {
      break; /* done !*/
    } else if (flags & BSC_TOKEN_PARSE_ALLOW_STORE) {
      bsc_token_ch_append(bsc, TOLOWER(ch));
    }
  }
  if (flags & BSC_TOKEN_PARSE_ALLOW_UNGET) {
    pfile_ch_unget(pf, ch);
  }
  return ch;
}



/*
 * NAME
 *   allow_*
 *
 * DESCRIPTION
 *   used by the parsing functions
 *
 * PARAMETERS
 *   ch : character to test
 *
 * RETURN
 *   TRUE if the character is allowed, FALSE if not
 *
 * NOTES
 *   none
 */
static boolean_t allow_identifier(int ch)
{
  return ISALPHA(ch) || ISDIGIT(ch) || ('_' == ch);
}

static boolean_t allow_decimal(int ch)
{
  return ISDIGIT(ch);
}

static boolean_t allow_hex(int ch)
{
  return ISXDIGIT(ch);
}

static boolean_t allow_binary(int ch)
{
  return ('0' == ch) || ('1' == ch);
}

static boolean_t allow_space(int ch)
{
  return ('\n' != ch) && ISSPACE(ch);
}

static boolean_t allow_comment(int ch)
{
  return ('\n' != ch);
}



/*
 * NAME
 *   bsc_token_get
 *
 * DESCRIPTION
 *   retrieve a token
 *
 * PARAMETERS
 *   pf    : returned by pfile_open()
 *   which : which token to get (first, current, next)
 *   dst   : [out] points to retrieved token
 *
 * RETURN
 *   none
 *
 * NOTES
 */
const char *bsc_token_get(bsc_file_t *bsc, bsc_token_get_t which)
{
  static const char xdig[] = "0123456789abcdef";
  int      ch;
  pfile_t *pf;

  pf = bsc_file_pf_get(bsc);

  switch (which) {
    case bsc_token_first:
      pfile_rewind(pf);
      /* fall through */
    case bsc_token_next:
      /* skip any preceding spaces */
      ch = bsc_token_parse_allow(bsc, allow_space, 0);
      bsc_token_reset(bsc);
      if (ISALPHA(ch)) {
        ch = TOLOWER(ch);
      }
      bsc_token_ch_append(bsc, ch);
      if (('_' == ch) || ISALPHA(ch)) {
        /* identifier */
        ch = bsc_token_parse_allow(bsc, allow_identifier, 
          BSC_TOKEN_PARSE_ALLOW_STORE
          | BSC_TOKEN_PARSE_ALLOW_UNGET);
      } else if (ISDIGIT(ch)) {
        /* decimal constant */
        ch = bsc_token_parse_allow(bsc, allow_decimal, 
          BSC_TOKEN_PARSE_ALLOW_STORE);
        if ('.' == ch) {
          bsc_token_ch_append(bsc, ch);
          /*pfile_token_get(pf, pfile_token_next, 0); */
          bsc_token_parse_allow(bsc, allow_decimal,
               BSC_TOKEN_PARSE_ALLOW_STORE | BSC_TOKEN_PARSE_ALLOW_UNGET);
        } else {
          pfile_ch_unget(pf, ch);
        }
      } else {
        switch (ch) {
          case '%': /* binary constant        */
            bsc_token_parse_allow(bsc, allow_binary, 
              BSC_TOKEN_PARSE_ALLOW_STORE | BSC_TOKEN_PARSE_ALLOW_UNGET);
            break;
          case '$': /* hex constant           */
            bsc_token_parse_allow(bsc, allow_hex, 
              BSC_TOKEN_PARSE_ALLOW_STORE | BSC_TOKEN_PARSE_ALLOW_UNGET);
            break;
          case '!': /* might be !! */
            ch = pfile_ch_get(pf);
            if ('!' == ch) {
              bsc_token_ch_append(bsc, ch);
            } else {
              pfile_ch_unget(pf, ch);
            }
            break;
          case '&': /* might be && */
            ch = pfile_ch_get(pf);
            if ('&' == ch) {
              bsc_token_ch_append(bsc, ch);
            } else {
              pfile_ch_unget(pf, ch);
            }
            break;
          case '|': /* might be || */
            ch = pfile_ch_get(pf);
            if ('|' == ch) {
              bsc_token_ch_append(bsc, ch);
            } else {
              pfile_ch_unget(pf, ch);
            }
            break;
          case '<': /* might be < or <= or <> */
            ch = pfile_ch_get(pf);
            if (('=' == ch) || ('>' == ch) || ('<' == ch)) {
              bsc_token_ch_append(bsc, ch);
            } else {
              pfile_ch_unget(pf, ch);
            }
            break;
          case '>': /* might be > or >=       */
            ch = pfile_ch_get(pf);
            if (('=' == ch) || ('>' == ch)) {
              bsc_token_ch_append(bsc, ch);
            } else {
              pfile_ch_unget(pf, ch);
            }
            break;
          case '\'': /* comment, skip to EOL */
            ch = bsc_token_parse_allow(bsc, allow_comment, 0);
            /* the next character must either be EOL or EOF */
            bsc_token_reset(bsc);
            bsc_token_ch_append(bsc, ch);
            break;
          case '\"': /* string constant, ends at EOL or " */
            {
              enum {
                ch_type_normal = 0, /* normal processing */
                ch_type_esc,        /* escape character */
                ch_type_hex,        /* hex processing   */
                ch_type_oct         /* oct processing */
              } mode;
              char val;     /* value if processing hex or oct */

              val = 0;
              mode = ch_type_normal;

              do {
                ch = pfile_ch_get(pf);
                switch (mode) {
                  case ch_type_normal:
                    if ('\\' == ch) {
                      mode = ch_type_esc;
                    } else {
                      if ((EOF != ch) && ('\n' != ch)) {
                        bsc_token_ch_append(bsc, ch);
                      }
                    }
                    break;
                  case ch_type_esc:
                    mode = ch_type_normal;
                    if (ISOCTDIGIT(ch)) {
                      mode = ch_type_oct;
                      val = (ch - '0');
                    } else {
                      switch (ch) {
                        case 'a': 
                          bsc_token_ch_append(bsc, '\a'); /* bell */
                          break; 
                        case 'b': 
                          bsc_token_ch_append(bsc, '\b'); /* bkpc */
                          break;
                        case 'f': 
                          bsc_token_ch_append(bsc, '\f'); /* ff */
                          break;
                        case 'n': 
                          bsc_token_ch_append(bsc, '\n'); /* lf */
                          break;
                        case 'r': 
                          bsc_token_ch_append(bsc, '\r'); /* cr */
                          break;
                        case 't': 
                          bsc_token_ch_append(bsc, '\t'); /* h.tab */
                          break;
                        case 'v': 
                          bsc_token_ch_append(bsc, '\v'); /* v.tab */
                          break;
                        case 'x': 
                          mode = ch_type_hex; /* 1 or 2 hex digits */
                          val  = 0;
                          break;
                        default:
                          if ((EOF != ch) && ('\n' != ch)) {
                            bsc_token_ch_append(bsc, ch);
                            ch = 0;
                          }
                      }
                    }
                    break;
                    
                  case ch_type_oct:
                    if (ISOCTDIGIT(ch)) {
                      val = val * 8 + (ch - '0');
                    } else {
                      bsc_token_ch_append(bsc, val);
                      mode = ch_type_normal;
                      pfile_ch_unget(pf, ch);
                      ch = 0;
                    }
                    break;
                  case ch_type_hex:
                    if (ISXDIGIT(ch)) {
                      val = val * 16 + (strchr(xdig, TOLOWER(ch)) - xdig);
                    } else {
                      bsc_token_ch_append(bsc, ch);
                      mode = ch_type_normal;
                      pfile_ch_unget(pf, ch);
                      ch = 0;
                    }
                    break;
                }
              } while ((EOF != ch) && ('\n' != ch) && ('"' != ch));
              if ('"' != ch) {
                pfile_ch_unget(pf, ch);
              }
            }
            break;
          default: /* don't know what this is, so it must be its own token */
            break;
        }
      }
      /* fall through */
    case bsc_token_current:
      break;
  }
  return bsc_token_ptr_get(bsc);
}

/*
 * NAME
 *   bsc_token_to_str
 *
 * DESCRIPTION
 *   convert the current token to a string variable
 *
 * PARAMETERS
 *   bsc  :
 *   plog :
 *   dst  :
 *
 * RETURN
 *
 * NOTES
 */
result_t bsc_token_to_str(bsc_file_t *bsc, pfile_log_t plog, value_t **dst)
{
  const char *ptr;
  pfile_t    *pf;
  result_t    rc;

  pf = bsc_file_pf_get(bsc);
  ptr = bsc_token_get(bsc, bsc_token_current);
  if ('"' == *ptr) {
    value_t    *val;

    val = pfile_string_find(pf, strlen(ptr) - 2, ptr + 1);
    if (val) {
      rc = result_ok;
    } else {
      size_t      n;
      char        lblname[64];
      char       *strname;
      label_t    *lbl;
      size_t      strsz;

      n = bsc_file_strid_next(bsc);
      sprintf(lblname, "_lookup__str%u_0", n);
      strname = lblname + 8;
      strsz   = strlen(strname) - 2;
      lbl = pfile_label_find(pf, pfile_log_none, lblname);
      if (lbl) {
        /* we're already been here, nothing more to do */
        val = pfile_value_find(pf, pfile_log_err, strsz, strname);
        label_release(lbl);
        rc = (val) ? result_ok : result_not_found;
      } else {
        rc = pfile_label_alloc(pf, lblname, 0, &lbl);
        if (result_ok == rc) {
          strname[strsz] = 0;

          strsz = strlen(ptr) - 2;
          rc = pfile_value_alloc(pf, strname, VARIABLE_FLAG_CONST,
            1, strsz, 0, &val);
          if (result_ok == rc) {
            size_t      ii;
            variable_t *var;

            var = value_variable_get(val);
            for (ii = strsz; ii; ii--) {
              /* strings are read backwards when printed, 
               * so record it backwards */
              variable_const_set(var, strsz - ii, ptr[ii]);
            }
            label_release(lbl);
          }
        }
      }
    }
    if (result_ok == rc) {
      *dst = val;
    }
  } else {
    pfile_log(pf, plog, BSC_MSG_STRING_EXPECTED);
    rc = result_invalid;
  }
  return rc;
}

/*
 * NAME
 *   bsc_token_to_constant
 *
 * DESCRIPTION
 *   convert the current token to an unnamed constant
 *
 * PARAMETERS
 *   pf   : returned by pfile_open()
 *   plog : how to log an invalid result
 *   dst  : [out] holds result on success
 *
 * RETURN
 *   0      : no error
 *   EINVAL : current token is not a constant
 *   ERANGE : constant is too large
 *
 * NOTES
 *   [dst] will be *locked* on successful return
 */
result_t bsc_token_to_constant(bsc_file_t *bsc, pfile_log_t plog,
    value_t **dst)
{
  int            base;
  const char    *ptr;
  unsigned long  n;
  result_t       rc;
  pfile_t       *pf;
  value_t       *val;

  pf = bsc_file_pf_get(bsc);
  ptr = bsc_token_get(bsc, bsc_token_current);
  if (ISDIGIT(*ptr)) {
    base = 10;
  } else if ('$' == *ptr) {
    ptr++;
    base = 16;
  } else if ('%' == *ptr) {
    ptr++;
    base = 2;
  } else {
    base = 0;
  }
  if (0 == base) {
    rc = result_invalid;
  } else {
    char *eptr;

    rc = result_ok;
    n = strtoul(ptr, &eptr, base);
    if ((ULONG_MAX == n) && (ERANGE == errno)) {
      rc = result_range;
    } else if (*eptr) {
      if (10 != base) {
        rc = result_invalid;
      } else {
        if ('.' == *eptr) {
          n = ceil(1000.0 * strtod(ptr, 0));
        }
      }
    }
    if (result_ok == rc) {
      rc = pfile_constant_get(pf, n, &val);
    }
  }
  if (result_invalid == rc) {
    pfile_log(pf, plog, BSC_MSG_CONSTANT_EXPECTED);
  } else if (result_range == rc) {
    pfile_log(pf, pfile_log_err, BSC_MSG_CONSTANT_RANGE);
  } else if (result_ok == rc) {
    *dst = val;
  }
  return rc;
}

const char *bsc_token_hack(bsc_file_t *bsc)
{
  int      ch;
  size_t   ct;
  pfile_t *pf;

  pf = bsc_file_pf_get(bsc);
  /* skip any spaces */
  do {
    ch = pfile_ch_get(pf);
  } while (('\n' != ch) && ISSPACE(ch));

  /* read the next word */
  bsc_token_reset(bsc);
  bsc_token_ch_append(bsc, ch);
  for (ct = 1; ; ct++) {
    ch = pfile_ch_get(pf);
    if ((EOF == ch) || ISSPACE(ch)) {
      pfile_ch_unget(pf, ch);
      break; /* <--- */
    }
    bsc_token_ch_append(bsc, TOLOWER(ch));
  }
  return bsc_token_ptr_get(bsc);
}

size_t bsc_varname_base_sz(const char *str)
{
  size_t sz;

  for (sz = strlen(str);
      ((sz > 2) 
        && ('_' == str[sz - 2]) 
        && (('h' == str[sz - 1]) || ('l' == str[sz - 1])));
      sz -= 2)
    ;
  return sz;
}

/*
 * NAME
 *   bsc_token_to_variable
 *
 * DESCRIPTION
 *   return the variable named by the current token
 *
 * PARAMETERS
 *   pf : returned by pfile_open
 *   plog : how to log result
 *   dst  : [out] holds result on success
 *
 * RETURN
 *   0      : no error
 *   EINVAL : token is not an identifier
 *   ENOENT : variable does not exist
 *
 * NOTES
 *   *dst is locked when returned
 */
result_t bsc_identifier_to_variable(bsc_file_t *bsc, pfile_log_t plog,
  const char *id, variable_t **dst)
{
  pfile_t    *pf;
  variable_t *var;
  result_t    rc;

  pf = bsc_file_pf_get(bsc);
  var = 0;
  if (('_' != *id) && !ISALPHA(*id)) {
    pfile_log(pf, plog, BSC_MSG_VARIABLE_EXPECTED);
    rc = result_invalid;
  } else {
    size_t         sz;

    sz = bsc_varname_base_sz(id);
    var = pfile_variable_find(pf, pfile_log_err, sz, id);
    if (var) {
      rc = result_ok;
      *dst = var;
    } else {
      rc = result_not_found;
    }
  }
  return rc;
}

result_t bsc_token_to_variable(bsc_file_t *bsc, pfile_log_t plog,
    variable_t **dst)
{
  return bsc_identifier_to_variable(bsc, plog,
    bsc_token_get(bsc, bsc_token_current), dst);
}

result_t bsc_identifier_to_value(bsc_file_t *bsc, pfile_log_t plog, 
  const char *id, value_t **dst)
{
  variable_t *var;
  value_t    *val;
  result_t    rc;

  rc = bsc_identifier_to_variable(bsc, plog, id, &var);
  if (result_ok == rc) {
    rc = value_alloc(&val, var, 0, 0, variable_sz_get(var));
    variable_release(var);
    if (result_ok != rc) {
      pfile_log_syserr(bsc_file_pf_get(bsc), rc);
    } else {
      id += strlen(variable_name_get(var));
      while ((result_ok == rc) && *id) {
        variable_sz_t sz;

        sz = value_sz_get(val) / 2;
        if (!sz) {
          pfile_log(bsc_file_pf_get(bsc), pfile_log_err, 
              BSC_MSG_ZERO_SZ);
          rc = result_invalid;
        } else if (0 == memcmp(id, "_l", 2)) {
          /* nothing to do */
        } else if (0 == memcmp(id, "_h", 2)) {
          value_baseofs_set(val, value_baseofs_get(val) + sz);
        } else {
          rc = result_internal;
        }
        value_sz_set(val, sz);
        id += 2;
      }
      if (result_ok == rc) {
        *dst = val;
      } else if (val) {
        value_release(val);
      }
    }
  }
  return rc;
}

result_t bsc_token_to_value(bsc_file_t *bsc, pfile_log_t plog, value_t **dst)
{
  return bsc_identifier_to_value(bsc, plog, 
    bsc_token_get(bsc, bsc_token_current), dst);
}

result_t bsc_token_to_label(bsc_file_t *bsc, pfile_log_t plog, 
    boolean_t alloc, label_t **dst)
{
  label_t    *lbl;
  const char *ptr;
  pfile_t    *pf;
  result_t    rc;

  pf = bsc_file_pf_get(bsc);
  ptr = bsc_token_get(bsc, bsc_token_current);
  if ((*ptr != '_') && !ISALPHA(*ptr)) {
    rc = result_invalid;
  } else {
    rc = result_ok;
    lbl = pfile_label_find(pf, (alloc) ? pfile_log_none : plog, ptr);
    if (!lbl && alloc) {
      rc = pfile_label_alloc(pf, ptr, 0, &lbl);
    }
    if (result_ok == rc) {
      if (!lbl) {
        rc = result_not_found;
      } else {
        *dst = lbl;
      }
    }
  }
  return rc;
}

boolean_t bsc_token_is(bsc_file_t *bsc, bsc_token_get_t which, 
    pfile_log_t plog, const char *str)
{
  int         rc;
  const char *ptr;
  pfile_t    *pf;

  pf = bsc_file_pf_get(bsc);
  ptr = bsc_token_get(bsc, which);
  rc = (0 == strcmp(ptr, str));
  if (!rc) {
    pfile_log(pf, plog, BSC_MSG_EXPECTED, str);
  }
  return rc;
}

boolean_t bsc_token_is_eof(bsc_file_t *bsc)
{
  const char *ptr;

  ptr = bsc_token_get(bsc, bsc_token_current);
  return (0 == *ptr);
}

boolean_t bsc_token_is_eos(bsc_file_t *bsc)
{
  return bsc_token_is_eof(bsc)
    || bsc_token_is(bsc, bsc_token_current, pfile_log_none, "\n");
}

