/**********************************************************
 **
 ** cmd.c : the cmd_type base functions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include <stdlib.h>
#include <string.h>

#include "variable.h"
#include "label.h"
#include "pfilesrc.h"
#include "cmd.h"

void *cmd_alloc(label_t *label, cmd_type_t type, 
  const cmd_vtbl_t *vtbl, size_t sz)
{
  cmd_t *cmd;

  cmd = MALLOC(sz);
  if (cmd) {
    cmd->link  = 0;
    cmd->flags = 0;
    cmd->src   = 0;
    cmd->line  = 0;
    cmd->label = 0;
    cmd->type  = type;
    cmd->vtbl  = vtbl;
    cmd_label_set(cmd, label);
  }
  return cmd;
}

void cmd_free(cmd_t *cmd)
{
  if (cmd->vtbl && cmd->vtbl->free_fn) {
    cmd->vtbl->free_fn(cmd);
  }
  cmd_label_set(cmd, 0);
  pfile_source_release(cmd->src);
  memset(cmd, 0xff, sizeof(*cmd));
  FREE(cmd);
}

void cmd_list_free(cmd_t *cmd)
{
  while (cmd) {
    cmd_t *link;

    link = cmd->link;
    cmd_free(cmd);
    cmd = link;
  }
}

void cmd_list_reverse(cmd_t **lst)
{
  cmd_t *pv;
  cmd_t *ptr;
  cmd_t *nxt;

  for (pv = 0, ptr = *lst;
       ptr;
       ptr = nxt)
  {
     nxt = ptr->link;
     ptr->link = pv;
     pv = ptr; 
     ptr = nxt;
  }
  *lst = pv;
}

cmd_t *cmd_link_get(const cmd_t *cmd)
{
  return (cmd) ? cmd->link : 0;
}

void         cmd_link_set(cmd_t *cmd, cmd_t *link)
{
  cmd->link = link;
}

cmd_type_t cmd_type_get(const cmd_t *cmd)
{
  return (cmd) ? cmd->type : cmd_type_nop;
}

void       cmd_type_set(cmd_t *cmd, cmd_type_t type)
{
  cmd->type = type;
}

unsigned   cmd_line_get(const cmd_t *cmd)
{
  return (cmd) ? cmd->line : -1;
}

void       cmd_line_set(cmd_t *cmd, unsigned line)
{
  cmd->line = line;
}

pfile_source_t *cmd_source_get(const cmd_t *cmd)
{
  return (cmd) ? cmd->src : 0;
}

void cmd_source_set(cmd_t *cmd, pfile_source_t *src)
{
  pfile_source_lock(src);
  if (cmd->src) {
    pfile_source_release(cmd->src);
  }
  cmd->src = src;
}

label_t *cmd_label_get(const cmd_t *cmd)
{
  return (cmd) ? cmd->label : 0;
}

void cmd_label_set(cmd_t *cmd, label_t *lbl)
{
  if (lbl) {
    lbl = label_master_get(lbl);
    label_lock(lbl);
    if ((cmd_type_branch == cmd_type_get(cmd))
      && (cmd_branchcond_none == cmd_brcond_get(cmd))) {
      label_brtype_set(lbl, cmd_brtype_get(cmd));
      label_brdest_set(lbl, cmd_brdst_get(cmd));
    }
  }
  if (cmd->label) {
    label_release(cmd->label);
  }
  cmd->label = lbl;
}

void cmd_dump(cmd_t *cmd, FILE *dst)
{
  if (cmd->label) { /*  && label_usage_get(cmd->label)) { */
    fprintf(dst, "%s: ", label_name_get(cmd->label));
  }
  if (cmd->vtbl && cmd->vtbl->dump_fn) {
    cmd->vtbl->dump_fn(cmd, dst);
  } else {
    const char *str;

    str = "{unknown}";
    switch (cmd->type) {
      case cmd_type_isr_cleanup: str = "{isr_cleanup}"; break;
      case cmd_type_end:         str = "{end}";         break;
      case cmd_type_sleep:       str = "{sleep}";       break;
      case cmd_type_nop:         str = "{nop}";         break;
      case cmd_type_clear:       str = "{clear}";       break;
      case cmd_type_branch:      str = "{branch}";      break;
      case cmd_type_intrinsic:   str = "{intrinsic}";   break;
      case cmd_type_operator:    str = "{operator}";    break;
    }
    fprintf(dst, str);
  }
  fputc('\n', dst);
}

void cmd_flag_set_all(cmd_t *cmd, flag_t flags)
{
  cmd->flags = flags;
}

void cmd_flag_set(cmd_t *cmd, flag_t flag)
{
  cmd->flags = (cmd->flags & ~flag) | flag;
}

void cmd_flag_clr(cmd_t *cmd, flag_t flag)
{
  cmd->flags = (cmd->flags & ~flag);
}

boolean_t cmd_flag_test(const cmd_t *cmd, flag_t flag)
{
  return (cmd) 
    ? ((cmd->flags & flag) == flag)
    : 0;
}



