/**********************************************************
 **
 ** label.c : manipulators for label_t
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include <stdlib.h>
#include <string.h>

#include "pfile.h"
#include "slist.h"
#include "label.h"
#include "lbl_decl.h"

static void label_free(slist_el_t *el)
{
  label_t *lbl;

  lbl = slist_el_data(el);
  if (lbl->free_cb) {
    lbl->free_cb(lbl);
  }
  FREE(lbl);
}


/* basic functions */
result_t label_alloc(label_t **dst, const char *name, 
  label_free_cb_t free_cb)
{
  label_t *lbl;
  int      rc;
  size_t   sz;

  sz = (name) ? 1 + strlen(name) : 0;

  lbl = MALLOC(sizeof(*lbl) + sz);
  if (!lbl) {
    rc = result_memory;
  } else {
    rc = result_ok;
    lbl->flags = 0;
    lbl->pc    = 0;
    lbl->data  = 0;
    lbl->free_cb = free_cb;
    lbl->master  = lbl;
    lbl->usage   = 0;
    lbl->brtype   = cmd_branchtype_none;
    lbl->brdest   = 0;
    if (name) {
      lbl->name = (void *) (lbl + 1);
      memcpy(lbl->name, name, sz);
    } else {
      lbl->name = 0;
    } 
    if (0 == rc) {
      slist_el_init(&lbl->el, label_free);
      *dst = lbl;
    }
  }
  return rc;
}

void label_release(label_t *lbl)
{
  if (lbl) {
    slist_el_release(&lbl->el);
  }
}

void label_lock(label_t *lbl)
{
  slist_el_lock(&lbl->el);
}

/* member retrieval */
const char *label_name_get(const label_t *lbl)
{
  return (lbl) ? lbl->name : 0;
}

slist_el_t *label_slist_el_get(label_t *lbl)
{
  return (lbl) ? &lbl->el : 0;
}

void label_list_init(lbllist_t *lst)
{
  slist_init(&lst->lst);
}

void label_list_append(lbllist_t *lst, label_t *lbl)
{
  slist_append(&lst->lst, &lbl->el);
}

label_t *label_list_find(lbllist_t *lst, lblfind_cb_t cb, void *cb_arg,
    const void *data)
{
  label_t *lbl;

  for (lbl = slist_el_data(slist_head(&lst->lst));
       lbl && !cb(cb_arg, lbl, data);
       lbl = slist_el_data(slist_el_next(&lbl->el)))
    ;
  if (lbl) {
    label_lock(lbl);
  }
  return lbl;
}

refct_t label_refct_get(const label_t *lbl)
{
  return (lbl) ? slist_el_refct_get(&lbl->el) : 0;
}

void label_list_reset(lbllist_t *lst, pfile_t *pf)
{
  while (slist_head(&lst->lst)) {
    if (label_refct_get(label_list_head(lst)) > 1) {
      pfile_log(pf, pfile_log_debug, PFILE_MSG_DBG_LABEL_LEAK, 
          label_name_get(label_list_head(lst)));
    }
    slist_el_remove(slist_head(&lst->lst));
  }
}

unsigned label_list_ct_get(const lbllist_t *lst)
{
  return (lst) ? slist_ct_get(&lst->lst) : 0;
}

void label_dump(const label_t *lbl, FILE *out)
{
  fprintf(out, "%s", label_name_get(lbl));
}

void *label_data_get(const label_t *lbl)
{
  return (lbl) ? lbl->data : 0;
}

void label_data_set(label_t *lbl, void *data)
{
  lbl->data = data;
}

void label_master_set(label_t *lbl, label_t *master)
{
  lbl->master = master;
}

label_t *label_master_get(const label_t *lbl)
{
  return (lbl) ? lbl->master : 0;
}

label_t *label_list_head(lbllist_t *lst)
{
  return slist_el_data(slist_head(&lst->lst));
}

label_t *label_next_get(label_t *lbl)
{
  return (lbl) ? slist_el_data(slist_el_next(&lbl->el)) : 0;
}

cmd_branchtype_t label_brtype_get(const label_t *lbl)
{
  return (lbl) ? lbl->brtype : cmd_branchtype_none;
}

void label_brtype_set(label_t *lbl, cmd_branchtype_t type)
{
  lbl->brtype = type;
}

label_t *label_brdest_get(const label_t *lbl)
{
  return (lbl) ? lbl->brdest : 0;
}

void label_brdest_set(label_t *lbl, label_t *brdest)
{
  lbl->brdest = brdest;
}

ctr_t label_usage_get(const label_t *lbl)
{
  return (lbl) ? lbl->usage : 0;
}

void label_usage_bump(label_t *lbl, ctr_bump_t chg)
{
  switch (chg) { 
    case ctr_bump_incr: 
      ++lbl->usage; 
      break;
    case ctr_bump_decr: 
      if (lbl->usage) {
        --lbl->usage; 
      }
      break;
  }
}

size_t label_pc_get(label_t *lbl)
{
  return (lbl) ? lbl->pc : 0;
}

void label_pc_set(label_t *lbl, size_t pc)
{
  lbl->pc = pc;
}

ctr_t label_ref_ct_get(const label_t *lbl)
{
  return (lbl) ? slist_el_refct_get(&lbl->el) : 0;
}

boolean_t label_flag_test(const label_t *lbl, flag_t flag)
{
  return (lbl) ? ((lbl->flags & flag) == flag): 0;
}

void label_flag_set(label_t *lbl,  flag_t flag)
{
  lbl->flags |= flag;
}

void label_remove(label_t *lbl)
{
  slist_el_remove(&lbl->el);
}

