/************************************************************
 **
 ** pic.c : pic code generation definitions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <stdlib.h>
#include <string.h>

#include "pfile.h"
#include "cmd.h"
#include "operator.h"
#include "bsc_cmd.h"
#include "pic_opfn.h"
#include "pic_op.h"
#include "picdelay.h"
#include "pic.h"
#include "pic_var.h"
#include "pic_daop.h"
#include "pic_brop.h"

#define LABEL_PIC_RESET "_pic_reset"
#define LABEL_PIC_ISR   "_pic_isr"
variable_sz_t mul_sz;
variable_sz_t div_sz;

pic_mem_size_t pic_mem_size;
pic_mem_size_t pic_mem_size_loader;

static const char *pic_opcode_str(pic_opcode_t op)
{
  const char *ostr;

  ostr = "{unknown}";
  switch (op) {
    case pic_opcode_none:   ostr = "{none}"; break;
    case pic_opcode_addwf:  ostr = "addwf";  break;
    case pic_opcode_andwf:  ostr = "andwf";  break;
    case pic_opcode_xorwf:  ostr = "xorwf";  break;
    case pic_opcode_iorwf:  ostr = "iorwf";  break;
    case pic_opcode_subwf:  ostr = "subwf";  break;
    case pic_opcode_comf:   ostr = "comf";   break;
    case pic_opcode_decf:   ostr = "decf";   break;
    case pic_opcode_decfsz: ostr = "decfsz"; break;
    case pic_opcode_incf:   ostr = "incf";   break;
    case pic_opcode_incfsz: ostr = "incfsz"; break;
    case pic_opcode_rlf:    ostr = "rlf";    break;
    case pic_opcode_rrf:    ostr = "rrf";    break;
    case pic_opcode_movf:   ostr = "movf";   break;
    case pic_opcode_swapf:  ostr = "swapf";  break;
    case pic_opcode_clrf:   ostr = "clrf";   break;
    case pic_opcode_clrw:   ostr = "clrw";   break;
    case pic_opcode_movwf:  ostr = "movwf";  break;
    case pic_opcode_nop:    ostr = "nop";    break;
    case pic_opcode_retfie: ostr = "retfie"; break;
    case pic_opcode_ret:    ostr = "return"; break;
    case pic_opcode_sleep:  ostr = "sleep";  break;
    case pic_opcode_clrwdt: ostr = "clrwdt"; break;
    case pic_opcode_bcf:    ostr = "bcf";    break;
    case pic_opcode_bsf:    ostr = "bsf";    break;
    case pic_opcode_btfsc:  ostr = "btfsc";  break;
    case pic_opcode_btfss:  ostr = "btfss";  break;
    case pic_opcode_addlw:  ostr = "addlw";  break;
    case pic_opcode_andlw:  ostr = "andlw";  break;
    case pic_opcode_iorlw:  ostr = "iorlw";  break;
    case pic_opcode_movlw:  ostr = "movlw";  break;
    case pic_opcode_sublw:  ostr = "sublw";  break;
    case pic_opcode_xorlw:  ostr = "xorlw";  break;
    case pic_opcode_call:   ostr = "call";   break;
    case pic_opcode_goto:   ostr = "goto";   break;
    case pic_opcode_org:    ostr = "org";    break;
    case pic_opcode_end:    ostr = "end";    break;
    case pic_opcode_retlw:  ostr = "retlw";  break;
    case pic_opcode_datalo_set: ostr = "datalo_set"; break;
    case pic_opcode_datalo_clr: ostr = "datalo_clr"; break;
    case pic_opcode_datahi_set: ostr = "datahi_set"; break;
    case pic_opcode_datahi_clr: ostr = "datahi_clr"; break;
    case pic_opcode_branchlo_set: ostr = "branchlo_set"; break;
    case pic_opcode_branchlo_clr: ostr = "branchlo_clr"; break;
    case pic_opcode_branchhi_set: ostr = "branchhi_set"; break;
    case pic_opcode_branchhi_clr: ostr = "branchhi_clr"; break;
  }
  return ostr;
}

pic_code_t *pic_code_lst; /* list of codes */

/************************************************************
 **
 ** PIC output functions
 **   I don't like ''x+0'' so I eliminate it!
 **
 ************************************************************/
pic_code_t *pic_code_alloc(void)
{
  pic_code_t *code;

  code = MALLOC(sizeof(*code));
  if (code) {
    memset(code, 0, sizeof(*code));
  }
  return code;
}

pic_code_t *pic_code_get(pfile_t *pf,
    label_t *lbl, pic_opcode_t op)
{
  pic_code_t *code;

  code = pic_code_lst;

  /* if (!code || (lbl && (pic_opcode_none != code->op))) { */
    code = pic_code_alloc();
    code->link = pic_code_lst;
    pic_code_lst = code;
  /* } */
  if (lbl) {
    label_lock(lbl);
    code->lbl = lbl;
  }
  code->op  = op;
  return code;
}

/************************************************************
 **
 ** utility functions
 **
 ************************************************************/
/*
 * NAME
 *   pic_accum_get
 *
 * DESCRIPTION
 *   return an accumulator setting its size and flags appropriately
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
static value_t *pic_internal_var_get(pfile_t *pf, const char *name,
  variable_sz_t sz, flag_t flags)
{
  value_t    *tmp;

  tmp = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF, name);
  if (tmp) {
    variable_t *tvar;

    tvar = value_variable_get(tmp);
    if (variable_sz_get(tvar) < sz) {
      variable_sz_set(tvar, sz);
    }
    value_sz_set(tmp, sz);
    if (flags & VARIABLE_FLAG_SIGNED) {
      value_flag_set(tmp, VARIABLE_FLAG_SIGNED);
    } else {
      value_flag_clr(tmp, VARIABLE_FLAG_SIGNED);
    }
  }
  return tmp;
}

value_t *pic_accum_get(pfile_t *pf, unsigned n, variable_sz_t sz,
    flag_t flags)
{
  char        name[16];

  sprintf(name, "_accum%u", n);
  return pic_internal_var_get(pf, name, sz, flags);
}

value_t *pic_sign_get(pfile_t *pf)
{
  return pic_internal_var_get(pf, "_sign", 1, 0);
}

value_t *pic_loop_get(pfile_t *pf)
{
  return pic_internal_var_get(pf, "_loop", 1, 0);
}

/*
 * NAME
 *   pic_branch
 *
 * DESCRIPTION
 *   handle command type cmd_type_branch
 *
 * PARAMETERS
 *   pf  : pfile
 *   cmd : a branch command
 *
 * RETURN
 *
 * NOTES
 */
void pic_branch(pfile_t *pf, cmd_branchtype_t brtype,
    cmd_branchcond_t brcond, label_t *dst, value_t *val)
{
  pic_opcode_t     op;

  if (cmd_branchcond_none != brcond) {
    /* val might be NULL if Z is already set elsewhere */
    if (val) {
      pic_op(pf, operator_logical, 0, val, 0);
      /* pic_instr_append_w_kn(pf, pic_opcode_andlw, 1); */
      op = pic_opcode_nop;
      switch (brcond) {
        case cmd_branchcond_true:  op = pic_opcode_btfss; break;
        case cmd_branchcond_false: op = pic_opcode_btfsc; break;
        case cmd_branchcond_none:  break;
      }
      pic_instr_append_reg_flag(pf, op, "_status", "_z");
    }
  }
  op = pic_opcode_nop;
  switch (brtype) {
    case cmd_branchtype_none:   break;
    case cmd_branchtype_goto:   op = pic_opcode_goto; break;
    case cmd_branchtype_call:   op = pic_opcode_call; break;
    case cmd_branchtype_return: op = pic_opcode_ret; break;
  }
  if (pic_opcode_ret == op) {
    pic_instr_append(pf, op);
  } else {
    pic_instr_append_n(pf, op, dst);
  } 
}

/*
 * NAME
 *   pic_isr_cleanup
 *
 * DESCRIPTION
 *   handle command type cmd_type_isr_cleanup
 *
 * PARAMETERS
 *   pf  : pfile
 *   cmd : command
 *
 * RETURN
 *   none
 *
 * NOTES
 */
static void pic_isr_cleanup(pfile_t *pf)
{
  value_t    *codegen;

  /* create the pic isr cleanup */
  codegen = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_codegen");
  if (codegen) {
    value_t    *isr_codegen;

    isr_codegen = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF, 
      "_isrcodegen");
    if (isr_codegen) {

      pic_op(pf, operator_assign, codegen, isr_codegen, 0);
      pic_instr_append_reg_d(pf, pic_opcode_movf, "_isr_pclath", pic_opdst_w);
      pic_instr_append_reg(pf, pic_opcode_movwf, "_pclath");
      pic_instr_append_reg_d(pf, pic_opcode_swapf, "_isr_status", pic_opdst_w);
      pic_instr_append_reg(pf, pic_opcode_movwf, "_status");
      pic_instr_append_reg_d(pf, pic_opcode_swapf, "_isr_w", pic_opdst_f);
      pic_instr_append_reg_d(pf, pic_opcode_swapf, "_isr_w", pic_opdst_w);
      pic_instr_append(pf, pic_opcode_retfie);

      value_release(isr_codegen);
    }
    value_release(codegen);
  }
}

/*
 * NAME
 *   pic_end
 *
 * DESCRIPTION
 *   handle commands of type cmd_type_end
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 *   this simply loops to a sleep instruction
 */
static void pic_end(pfile_t *pf)
{
  label_t *lbl;

  if (result_ok == pfile_label_alloc(pf, 0, 0, &lbl)) {
    pic_instr_append_label(pf, lbl);
    pic_instr_append(pf, pic_opcode_sleep);
    pic_instr_append_n(pf, pic_opcode_goto, lbl);
    label_release(lbl);
  }
}

/*
 * NAME
 *   pic_clear
 *
 * DESCRIPTION
 *   set all data to 0
 *
 * PARAMETERS
 *   none
 *
 * RETURN
 *
 * NOTES
 */
static void pic_clear(pfile_t *pf)
{
  pic_bank_info_t *bnk;
  size_t           bnk_no;

  for (bnk = pic_bank_info_get(pf), bnk_no = 0; bnk; bnk = bnk->link, bnk_no++) {
    /* if there's nothing allocated in this bank, do nothing */
    size_t sz;

    sz = bnk->next - bnk->lo;
    if (sz > 8) {
      /* loop */
      label_t *lbl;

      pic_instr_append_w_kn(pf, pic_opcode_movlw, sz);
      pic_instr_append_reg(pf, pic_opcode_movwf, "_loop");
      pic_instr_append_w_kn(pf, pic_opcode_movlw, bnk->lo);
      pic_instr_append_reg(pf, pic_opcode_movwf, "_fsr");
      pic_instr_append_reg_flag(pf, 
          (bnk->lo >= 256) ? pic_opcode_bsf : pic_opcode_bcf,
          "_status", "_irp");
      if (result_ok == pfile_label_alloc(pf, 0, 0, &lbl)) {
        pic_instr_append_label(pf, lbl);
        pic_instr_append_reg(pf, pic_opcode_clrf, "_ind");
        pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
        pic_instr_append_reg_d(pf, pic_opcode_decfsz, "_loop", pic_opdst_f);
        pic_instr_append_n(pf, pic_opcode_goto, lbl);
        label_release(lbl);
      }
    } else if (sz) {
      /* simply clear everything here */
      size_t pos;

      value_t *val;

      val = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_memory");
      if (val) {
        for (pos = bnk->lo; sz; pos++, sz--) {
          pic_instr_append_f(pf, pic_opcode_clrf, val, pos);
        }
        value_release(val);
      }
    }
  }
}

/*
 * NAME
 *   pic_branch_bittest_out
 *
 * DESCRIPTION
 *   output a bittest for a branch
 *
 * PARAMETERS
 *   pf   :
 *   val1 : value to test
 *   val2 : bit #
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this is similar to pic_bittest() but that cannot be used because
 *   it always assumes btfsc, whereas we need either btfsc or btfss
 *   here
 */
static void pic_brtest_out(pfile_t *pf, value_t *val1, value_t *val2,
    cmd_branchcond_t cond)
{
  variable_const_t n;
  pic_opcode_t     pop;

  pop = (cmd_branchcond_true == cond) ? pic_opcode_btfsc : pic_opcode_btfss;

  n = value_const_get(val2);
  if (n >= 8U * value_sz_get(val1)) {
    pfile_log(pf, pfile_log_warn, PIC_MSG_CONSTANT_RANGE);
    pic_instr_append(pf, pic_opcode_clrw);
    pic_instr_append_reg_flag(pf, pop, "_status", "_z");
  } else {
    if ((n / 8) || !value_name_get(val1)) {
      pic_instr_append_f_bn(pf, pop, val1, (size_t) n / 8, n & 7);
    } else {
      pic_instr_append_f_b(pf, pop, val1, (size_t) n / 8, val2);
    }
  }
}


/*
 * NAME
 *   pic_cmd_out
 *
 * DESCRIPTION
 *   convert a cmd_t to one or more PIC instructions
 *
 * PARAMETERS
 *   pf  : pfile
 *   cmd : cmd to convert
 *
 * RETURN
 *   none
 *
 * NOTES
 */
void pic_cmd_out(pfile_t *pf, cmd_t **cmd_ptr)
{
  pic_code_t *plst;
  pic_code_t *cptr;
  cmd_t      *cmd_nxt;
  cmd_t      *cmd;
  value_t    *val;

  cmd = *cmd_ptr;
  plst = pic_code_lst;
  if (cmd->label) {
    pic_instr_append_label(pf, cmd_label_get(cmd));
  }
  cmd_nxt = cmd_link_get(cmd);
  val = cmd_opval2_get(cmd);
  if (!val) {
    val = cmd_opval1_get(cmd);
  }
  if (value_is_const(val)
      && ((operator_bittst == cmd_optype_get(cmd)) 
        || ((operator_andb == cmd_optype_get(cmd))
          && !(value_const_get(val) & (value_const_get(val) - 1))))
      && (cmd_type_branch == cmd_type_get(cmd_nxt))
      && (0 == memcmp(value_name_get(cmd_opdst_get(cmd)), "_t", 2))
      && (value_same(cmd_opdst_get(cmd), cmd_brval_get(cmd_nxt)))) {
    /* this is a special case:
     *   bittst [tmp], [x], [constant bit]
     *   branch [true|false] [tmp] [dst]
     *   ..or..
     *   andb [tmp], [x], [constant bit]
     *   branch [true|false] [tmp] [dst]
     *
     * since andb x,0x80  --> bittst x,7
     */
    if (operator_bittst == cmd_optype_get(cmd)) {
      pic_brtest_out(pf, cmd_opval1_get(cmd), cmd_opval2_get(cmd),
          cmd_brcond_get(cmd_nxt));
      pic_branch(pf, cmd_brtype_get(cmd_nxt), cmd_branchcond_none,
          cmd_brdst_get(cmd_nxt), 0);
    } else {
      variable_const_t n;
      variable_const_t bit;

      n = value_const_get(val);
      for (bit = 0;
           0 != (n >>= 1);
           bit++)
        ; /* null body */
      if (result_ok == pfile_constant_get(pf, bit, &val)) {
        pic_brtest_out(pf, cmd_opval1_get(cmd), val,
            cmd_brcond_get(cmd_nxt));
        pic_branch(pf, cmd_brtype_get(cmd_nxt), cmd_branchcond_none,
            cmd_brdst_get(cmd_nxt), 0);
        value_release(val);
      }
    }
    *cmd_ptr = cmd_nxt;
    cmd = cmd_nxt;
  } else {
    switch (cmd_type_get(cmd)) {
      case cmd_type_operator:      
        pic_op(pf, cmd_optype_get(cmd), cmd_opdst_get(cmd), 
            cmd_opval1_get(cmd), cmd_opval2_get(cmd)); 
        break;
      case cmd_type_branch:    
        pic_branch(pf, cmd_brtype_get(cmd), cmd_brcond_get(cmd),
            cmd_brdst_get(cmd), cmd_brval_get(cmd)); 
        break;
      case cmd_type_nop:       
        pic_instr_append(pf, pic_opcode_nop);
        break;
      case cmd_type_sleep:
        pic_instr_append(pf, pic_opcode_sleep);
        break;
      case cmd_type_end:
        pic_end(pf);
        break;
      case cmd_type_intrinsic: 
        /* delay is handled at this level because there is no choice */
        if (bsc_intrinsic_type_delay == bsc_cmd_type_get(cmd)) {
          pic_delay_create(pf, value_const_get(
                bsc_cmd_param_get(cmd, 0)->u.var));
        }
        break;
      case cmd_type_isr_cleanup:
        pic_isr_cleanup(pf);
        break;
      case cmd_type_clear:
        pic_clear(pf);
        break;
    }
  }
  for (cptr = pic_code_lst;
       cptr && (cptr != plst);
       cptr = cptr->link) {
    cptr->cmd = cmd;
  }
}

/*
 * NAME
 *   pic_bitstate_changed
 *
 * DESCRIPTION
 *   determine if a bitstate has changed
 *
 * PARAMETERS
 *   old : original state
 *   new : new state
 *
 * RETURN
 *   1 : changed
 *   0 : no change
 *
 * NOTES
 *   a bit has changed if it's not the same as new and it's not
 *   indeterminte
 */

boolean_t pic_bitstate_changed(pic_bitstate_t old, pic_bitstate_t new)
{
  return ((old != new) 
      && (pic_bitstate_indeterminate != old));
}


/*
 * NAME
 *   pic_code_label_find
 *
 * DESCRIPTION
 *   find a label in the code
 *
 * PARAMETERS
 *   pf   : 
 *   lbl  : label to find
 *   ppcl : [out] holds result on success
 *
 * RETURN
 *   0      : no error
 *   ENOENT : label not found
 *
 * NOTES
 */
pic_code_t *pic_code_label_find(pfile_t *pf, const label_t *lbl)
{
  pic_code_t *ptr;

  if (lbl) {
    for (ptr = pic_code_lst; ptr && (ptr->lbl != lbl); ptr = ptr->link)
      ;
  } else {
    ptr = 0;
  }
  return ptr;
}

/*
 * NAME
 *   pic_cmd_dump
 *
 * DESCRIPTION
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
static void pic_bitstate_dump(pfile_t *pf, pfile_write_t where,
    pic_bitstate_t st, char set, char clr)
{
  char ch;

  ch = '!';
  switch (st) {
    case pic_bitstate_unknown:       ch = '-'; break;
    case pic_bitstate_set:           ch = set; break;
    case pic_bitstate_clr:           ch = clr; break;
    case pic_bitstate_indeterminate: ch = '?'; break;
  }
  pfile_write(pf, where, "%c", ch);
}


static void pic_code_dump(pfile_t *pf, const pic_code_t *code)
{
  const char *comma;
  int         sz;
  unsigned    pcode;

  pcode = pic_code_to_pcode(pf, code);
  if (0xffff != pcode) {
    pfile_write_hex(pf, 2 * code->pc, pcode & 0xff);
    pfile_write_hex(pf, 2 * code->pc + 1, (pcode >> 8) & 0xff);
  }

#if 0
  sz = pfile_write(pf, pfile_write_lst, "%04x %-16s", code->pc, 
      (code->lbl) ? label_name_get(code->lbl) : "");
#else
  sz = pfile_write(pf, pfile_write_asm, "%-32s",
      (code->lbl) ? label_name_get(code->lbl) : "");
#endif
  if (pic_opcode_none != code->op) {
    sz += pfile_write(pf, pfile_write_asm, " %-8s", pic_opcode_str(code->op));
    comma = "";
    if (code->val) {
      if (ISARRAY(code->val)) {
        /* this is an array, so use the _ind register */
        sz += pfile_write(pf, pfile_write_asm, "_ind");
      } else {
        size_t ofs;

        ofs = pic_var_ofs_get(pf, code->val, code->ofs);
        sz += pfile_write(pf, pfile_write_asm, "%s", value_name_get(code->val));
        if (ofs) {
          sz += pfile_write(pf, pfile_write_asm, "+%u", ofs);
        }
      }
      if (pic_opdst_none != code->dst) {
        sz += pfile_write(pf, pfile_write_asm, ",%c", 
            (pic_opdst_w == code->dst) ? 'w' : 'f');
      }
      comma = ",";
    }
    if (code->literal) {
      if (value_name_get(code->literal)) {
        sz += pfile_write(pf, pfile_write_asm, "%s%s", comma, 
            value_name_get(code->literal));
      } else if (pic_opcode_org != code->op) {
        sz += pfile_write(pf, pfile_write_asm, "%s%lu", comma, 
            value_const_get(code->literal) & 0xff);
      } else {
        sz += pfile_write(pf, pfile_write_asm, "%s%lu", comma, 
            value_const_get(code->literal));
      }
    }
    if (code->brdst) {
      sz += pfile_write(pf, pfile_write_asm, "%s%s", 
          (pic_opcode_movlw == code->op) ? "HIGH " : "", label_name_get(code->brdst));
    }
    if (pfile_flag_test(pf, PFILE_FLAG_DEBUG)) {
      if (sz < 60) {
        pfile_write(pf, pfile_write_asm, "%*s", 60 - sz, "");
      }
      pfile_write(pf, pfile_write_asm, ";");
      pic_bitstate_dump(pf, pfile_write_asm, code->databits.before.irp, 'I', 'i');
      pic_bitstate_dump(pf, pfile_write_asm, code->databits.before.rp1, 'R', 'r');
      pic_bitstate_dump(pf, pfile_write_asm, code->databits.before.rp0, 'S', 's');
      pfile_write(pf, pfile_write_asm, " ");
      pic_bitstate_dump(pf, pfile_write_asm, code->databits.action.irp, 'I', 'i');
      pic_bitstate_dump(pf, pfile_write_asm, code->databits.action.rp1, 'R', 'r');
      pic_bitstate_dump(pf, pfile_write_asm, code->databits.action.rp0, 'S', 's');
      pfile_write(pf, pfile_write_asm, " [");
      pic_bitstate_dump(pf, pfile_write_asm, code->branchbits.before.pclath4, 
          'H', 'h');
      pic_bitstate_dump(pf, pfile_write_asm, code->branchbits.before.pclath3, 
          'L', 'l');
      pfile_write(pf, pfile_write_asm, " ");
      pic_bitstate_dump(pf, pfile_write_asm, code->branchbits.action.pclath4, 
          'H', 'h');
      pic_bitstate_dump(pf, pfile_write_asm, code->branchbits.action.pclath3, 
          'L', 'l');
      pfile_write(pf, pfile_write_asm, "]");
#if 1
      pfile_write(pf, pfile_write_asm, " %04x %02x%02x", code->pc * 2, pcode & 0xff,
          pcode >> 8);
#else
      pfile_write(pf, pfile_write_asm, " %04x", code->pc);
#endif
    }
  }
  pfile_write(pf, pfile_write_asm, "\n");
}

void pic_cmd_dump(pfile_t *pf, const cmd_t *cmd, boolean_t first)
{
  pic_code_t *code;

  /* if first is TRUE, we need to dump the preamble */
  for (code = pic_code_lst; code && (code->cmd != cmd); code = code->link) {
    if (first) {
      pic_code_dump(pf, code);
    }
  }
  while (code && (code->cmd == cmd)) {
    pic_code_dump(pf, code);
    code = code->link;
  }
}

/*
 * NAME
 *   pic_code_lookup
 *
 * DESCRIPTION
 *   generate the lookup tables
 *
 * PARAMETERS
 *   pf  : pfile
 *   ct  : # of variable lists
 *   lst : variable list
 *   pc  : PC on entry
 *
 * RETURN
 *   none
 *
 * NOTES
 *   the lookup tables are at the start of the code because no single table
 *   can span 256-byte blocks. If put at the end, then ever time one of the
 *   tables is moved the branch optimization would need to be re-run
 */
typedef struct pic_lookup_sort_element_ {
  const variable_t *var;
  size_t            ofs; /* which byte   */
} pic_lookup_sort_element_t;

typedef struct pic_lookup_sort_ {
  size_t                     ct;
  pic_lookup_sort_element_t *el;
} pic_lookup_sort_t;

typedef void (*pic_code_lookup_cb_t)(void *arg, pfile_t *pf, const variable_t *var,
  size_t byte, label_t *lbl);

static void pic_code_lookup_iterate(pfile_t *pf, size_t ct, variable_t **vars, 
  pic_code_lookup_cb_t cb, void *cb_arg)
{
  size_t ii;

  /* create any lookup tables */
  /* first, determine how many there are; for multi-byte constant arrays
     we'll need one table per byte in constant's size. finally, if all bytes in
     a particular constant are the same, we don't bother allocating one */
  for (ii = 0; ii < ct; ii++) {
    const variable_t *vptr;

    for (vptr = vars[ii]; vptr; vptr = variable_link_get(vptr)) {
      if (variable_flag_test(vptr, VARIABLE_FLAG_CONST)
        && (variable_ct_get(vptr) > 1)) {
        if (variable_flag_test(vptr, VARIABLE_FLAG_USED)) {
          char     lbl_name[256];
          size_t   jj;

          for (jj = 0; jj < variable_sz_get(vptr); jj++) {
            label_t *lbl;

            sprintf(lbl_name, "_lookup_%s_%u", variable_name_get(vptr), jj);
            lbl = pfile_label_find(pf, pfile_log_none, lbl_name);
            if (lbl) {
              cb(cb_arg, pf, vptr, jj, lbl);
              label_release(lbl);
            }
          }
        }
      }
    }
  }
}

/* first iteration simply counts the # of tables that will be created */
static void pic_code_lookup_count_cb(void *arg, pfile_t *pf, const variable_t *var,
  size_t byte, label_t *lbl)
{
  (*(size_t *) arg)++;
}

/* this fills in the pic_label_sort_t structure */
static void pic_code_lookup_alloc_cb(void *arg, pfile_t *pf, const variable_t *var,
  size_t byte, label_t *lbl)
{
  pic_lookup_sort_t *srt;

  srt = arg;
  srt->el[srt->ct].var = var;
  srt->el[srt->ct].ofs = byte;
  srt->ct++;
}

static int pic_lookup_sort_el_cmp(const void *A, const void *B)
{
  const pic_lookup_sort_element_t *a = A;
  const pic_lookup_sort_element_t *b = B;
  variable_ct_t ct_a;
  variable_ct_t ct_b;
  int           cmp;

  ct_a = variable_ct_get(a->var);
  ct_b = variable_ct_get(b->var);

  if (ct_a > ct_b) {
    cmp = -1;
  } else if (ct_a == ct_b) {
    cmp = 0;
  } else {
    cmp = 1;
  }
  return cmp;
}

static void pic_code_lookup_create(pfile_t *pf, size_t ct, variable_t **vars,
  size_t pc)
{
  size_t total; /* # of lookup tables */


  total = 0;
  pic_code_lookup_iterate(pf, ct, vars, pic_code_lookup_count_cb, &total);
  if (total) {
    pic_lookup_sort_t srt;

    srt.ct = 0;
    srt.el = CALLOC(total, sizeof(*srt.el));
    if (!srt.el) {
      pfile_log_syserr(pf, result_memory);
    } else {
      size_t ii;

      pic_code_lookup_iterate(pf, ct, vars, pic_code_lookup_alloc_cb, &srt);
      qsort(srt.el, srt.ct, sizeof(*srt.el), pic_lookup_sort_el_cmp);

      for (ii = 0; ii < srt.ct; ii++) {
        if (srt.el[ii].var) {
          char              lbl_name[256];
          label_t          *lbl;
          const variable_t *vptr;
          size_t            jj;

          /* look for a table that will fit at the current location, if
             none exists we'll simply use the next table & change the PC */
          for (jj = ii; jj < srt.ct; jj++) {
            unsigned      b1;
            unsigned      b2;

            vptr = srt.el[jj].var;
            if (vptr) {
              b1 = pc / 256;
              b2 = (pc + variable_ct_get(vptr)) / 256;
              if (b1 == b2) {
                break;
              }
            }
          }
          if (jj == srt.ct) {
            jj = ii;
          }

          vptr = srt.el[jj].var;

          sprintf(lbl_name, "_lookup_%s_%u", variable_name_get(vptr), 
            srt.el[jj].ofs);
          lbl = pfile_label_find(pf, pfile_log_none, lbl_name);
          if (lbl) {
            variable_ct_t kk;
            unsigned      b1;
            unsigned      b2;

            b1 = pc / 256;
            b2 = (pc + variable_ct_get(vptr)) / 256;
            if (b1 != b2) {
              pc = (b1 + 1) * 256;
              pic_instr_append_w_kn(pf, pic_opcode_org, pc);
            }
            pic_instr_append_label(pf, lbl);
            pic_instr_append_reg_d(pf, pic_opcode_addwf, "_pcl", pic_opdst_f);
            for (kk = 0; kk < variable_ct_get(srt.el[jj].var); kk++) {
              pic_instr_append_w_kn(pf, pic_opcode_retlw, 
                (variable_const_get(vptr, kk) >> (8 * srt.el[jj].ofs)) & 0xff);
            }
            label_release(lbl);
            pc += (1 + variable_ct_get(vptr));
          }
          srt.el[jj].var = 0;
          if (jj != ii) {
            ii--;
          }
        }
      }
      FREE(srt.el);
    }
  }
}

/*
 * NAME
 *   pic_code_isr_preamble
 *
 * DESCRIPTION
 *   create the isr preamble code
 *
 * PARAMETERS
 *   pf : pfile
 *
 * RETURN
 *
 * NOTES
 *   from the PIC reference:
 *      movwf w_temp     ; w --> w_temp
 *      swapf status, w  ; status --> w (nybble swapped but not flags
 *                                      effected)
 *      clrf  status     ; 0 --> status (irp/rp1/rp0 = clr)
 *      movwf status_tmp ; w --> status_temp
 *      movf  pclath, w  ; pclath --> w
 *      movwf pclath_tmp ; w --> pclath_temp
 *      clrf  pclath     ; 0 --> pclath
 *      ...
 *      {isr}
 *      ...
 *      movf  pclath_tmp,w ; pclath_tmp --> w
 *      movwf pclath       ; w --> pclath
 *      swapf status_tmp,w ; status_tmp --> w
 *      movwf status       ; w --> status
 *      swapf w_temp,f     ; w_temp->h <--> w_temp->l --> w_temp
 *      swapf w_temp,w     ; w_temp->h <--> w_temp->l --> w
 *      retfie
 *
 *  also, the lookup tables may not cross 256 byte boundaries (giving a maximum of
 *    254 entries/table). Since I might need to futz with the ORGs here, I don't
 *    want to have to deal with moving them around if branchbits get inserted, so
 *    I'm going to waste a bit of space here & force the branchbits. it doesn't matter
 *    what I use as long as the order is:  HI, LO, goto. The analyzer will make
 *    any necessary corrections later.
 */
static void pic_code_preamble(pfile_t *pf, size_t ct, variable_t **vars)
{
  label_t *lbl_user;
  label_t *lbl_pic_reset; /* the reset vector */
  label_t *lbl_pic_isr;   /* the ISR vector */

  lbl_pic_reset = pfile_label_find(pf, pfile_log_none, LABEL_PIC_RESET);
  lbl_pic_isr   = pfile_label_find(pf, pfile_log_none, LABEL_PIC_ISR);

  if ((!lbl_pic_reset 
        && (result_ok != pfile_label_alloc(pf, LABEL_PIC_RESET, 0, &lbl_pic_reset)))
      || (!lbl_pic_isr
        && (result_ok != pfile_label_alloc(pf, LABEL_PIC_ISR, 0, &lbl_pic_isr)))) {
    if (lbl_pic_isr) {
      label_release(lbl_pic_isr);
    }
    if (lbl_pic_reset) {
      label_release(lbl_pic_reset);
    }
    return;
  }

  lbl_user = pfile_user_entry_get(pf);


  if (!lbl_user) {
    pfile_log(pf, pfile_log_crit, PIC_MSG_NO_USER_ENTRY);
  } else {
    label_t *lbl_isr;
    unsigned pc;

    lbl_isr  = pfile_isr_entry_get(pf);
    
    pc = (pfile_flag_test(pf, PFILE_FLAG_NORICK) ? 0 : 3);
    pic_instr_append_w_kn(pf, pic_opcode_org, pc);
    pic_instr_append_label(pf, lbl_pic_reset);
    if (!lbl_isr) {
      pic_instr_append_n(pf, pic_opcode_goto, lbl_user);
      pc += 1;
      pic_code_lookup_create(pf, ct, vars, pc);
    } else {
      /* nb: for the RF PIC loader, the ISR preamble code cannot exceed 2K */
      label_t *dummy_user;

      dummy_user = 0;
      if (pfile_flag_test(pf, PFILE_FLAG_NORICK)
            || (result_ok == pfile_label_alloc(pf, 0, 0, &dummy_user))) {
        pic_code_t *code;
        value_t    *isr_codegen;
        value_t    *codegen;

        if (dummy_user) {
          pic_instr_append_n(pf, pic_opcode_goto, dummy_user);
        } else {
          pic_instr_append(pf, pic_opcode_branchhi_set);
          pic_instr_append(pf, pic_opcode_branchlo_set);
          pic_instr_append_n(pf, pic_opcode_goto, lbl_user);
        }
        /* the interrupt vector is at 4, jump to the real ISR */
        pc = 4;
        pic_instr_append_w_kn(pf, pic_opcode_org, 4);
        pic_instr_append_label(pf, lbl_pic_isr);
        /* finally, the ISR preamble */
        code = pic_code_lst;
        pic_instr_append_reg(pf, pic_opcode_movwf, "_isr_w");
        pic_instr_append_reg_d(pf, pic_opcode_swapf, "_status", pic_opdst_w);
        pic_instr_append_reg(pf, pic_opcode_clrf, "_status");
        pic_instr_append_reg(pf, pic_opcode_movwf, "_isr_status");
        pic_instr_append_reg_d(pf, pic_opcode_movf, "_pclath", pic_opdst_w);
        pic_instr_append_reg(pf, pic_opcode_movwf, "_isr_pclath");
        pic_instr_append_reg(pf, pic_opcode_clrf, "_pclath");
        codegen = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF,
          "_codegen");
        if (codegen) {
          isr_codegen = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF,
            "_isrcodegen");
          if (isr_codegen) {
            pic_op(pf, operator_assign, isr_codegen, codegen, 0);
            value_release(isr_codegen);
          }
          value_release(codegen);
        }
        pic_instr_append(pf, pic_opcode_branchhi_set);
        pic_instr_append(pf, pic_opcode_branchlo_set);
        pic_instr_append_n(pf, pic_opcode_goto, pfile_isr_preamble_entry_get(pf));
        if (dummy_user) {
          /* jump to the real main */
          pic_instr_append_label(pf, dummy_user);
          pic_instr_append(pf, pic_opcode_branchhi_set);
          pic_instr_append(pf, pic_opcode_branchlo_set);
          pic_instr_append_n(pf, pic_opcode_goto, lbl_user);
          /* pc += 4; */
        }
        {
          pic_code_t *cptr;

          for (cptr = pic_code_lst; cptr && (cptr != code); cptr = cptr->link) {
            if (pic_opcode_none != cptr->op) {
              pc++;
            }
          }
        }
#if 0
        while (code) {
          pc++;
          code = code->link;
        }
#endif
#if 0
        /* put in any lookup tables */
        pic_instr_append_w_kn(pf, pic_opcode_org, 249);
        pic_code_lookup(pf, ct, vars, 249);
#else
        pic_instr_append_w_kn(pf, pic_opcode_org, pc);
        pic_code_lookup_create(pf, ct, vars, pc);
#endif
      }
      if (dummy_user) {
        label_release(dummy_user);
      }
      label_release(lbl_isr);
    }
    label_release(lbl_user);
  }
  label_release(lbl_pic_reset);
  label_release(lbl_pic_isr);
}


void pic_code_reset(pfile_t *pf)
{
  while (pic_code_lst) {
    pic_code_t *code;

    code = pic_code_lst;
    pic_code_lst = code->link;

    if (code->lbl) {
      label_release(code->lbl);
    }
    if (code->val) {
      value_release(code->val);
    }
    if (code->literal) {
      value_release(code->literal);
    }
    if (code->brdst) {
      label_release(code->brdst);
    }
    FREE(code);
  }
}

void pic_code_free(pfile_t *pf)
{
  if (!pfile_errct_get(pf)) {
    const pic_code_t *ptr;
    size_t            mem_max;

    /* the PIC loader requires 2K at the top of memory */
    mem_max = (pfile_flag_test(pf, PFILE_FLAG_NORICK)) 
      ? pic_mem_size
      : pic_mem_size_loader;

    for (ptr = pic_code_lst; ptr && ptr->link; ptr = ptr->link)
      ;
    pic_blist_info_log(pf);
    if (ptr) {
      pfile_log(pf, pfile_log_info, PIC_MSG_CODE_USED,
        (ptr) ? ptr->pc : 0, mem_max);
      if (ptr->pc >= mem_max) {
        pfile_log(pf, pfile_log_err, PIC_MSG_CODE_TOO_BIG);
      }
    }
  }
  pic_code_reset(pf);
  pic_blist_free(pf);
}

/*
 * NAME
 *   pic_cmd_alloc
 *
 * DESCRIPTION
 *   populate the pic_codes[] array
 *
 * PARAMETERS
 *   pf  : 
 *   cmd : first command
 *
 * RETURN
 *   none
 *
 * NOTES
 */
void pic_cmd_alloc(pfile_t *pf, cmd_t *cmd, size_t ct, variable_t **vars)
{
  /* dummy up _div32 & _mul32 */
  label_t *lbl_tmp;
  size_t   pcl;
  cmd_t   *cmd_ptr;
  unsigned pass;
  size_t   ii;
  pic_code_t *code;
  size_t ins_ct;
  value_t *printval;

  /* let's determine mul_sz and div_sz */
  pfile_log(pf, pfile_log_info, PIC_MSG_FINDING_INTRIN);
  for (cmd_ptr = cmd; cmd_ptr; cmd_ptr = cmd_link_get(cmd_ptr)) {
    if (cmd_type_operator == cmd_type_get(cmd_ptr)) {
      if ((operator_mul == cmd_optype_get(cmd_ptr))
        || (operator_div == cmd_optype_get(cmd_ptr))) {
        int              is_signed;
        int              is_shift;
        variable_sz_t    sz;

        value_t *val1;
        value_t *val2;

        if (cmd_opval2_get(cmd_ptr)) {
          val2 = cmd_opval2_get(cmd_ptr);
          val1 = cmd_opval1_get(cmd_ptr);
        } else {
          val2 = cmd_opval1_get(cmd_ptr);
          val1 = cmd_opdst_get(cmd_ptr);
        }

        sz = value_sz_get(val1);
        if (ISCONST(val2)) {
          variable_const_t n;

          is_signed = ISSIGNED(val1);
          n = value_const_get(val2);
          is_shift = !(n & (n - 1));
        } else {
          is_signed = ISSIGNED(val1) && ISSIGNED(val2);
          is_shift  = 0;
        }
        if (sz < value_sz_get(val2)) {
          sz = value_sz_get(val2);
        }
        if (is_signed || !is_shift) {
          if (operator_mul == cmd_optype_get(cmd_ptr)) {
            if (sz > mul_sz) {
              mul_sz = sz;
            }
          } else {
            if (sz > div_sz) {
              div_sz =sz;
            }
          }
        }
      }
    }
  }

  /* if printval is used, we might need to adjust the divide size */
  printval = pfile_value_find(pf, pfile_log_none, PFILE_LENGTH_DEF,
      "_printval");
  if (printval) {
    if (value_sz_get(printval) > div_sz) {
      div_sz = value_sz_get(printval);
    }
    value_release(printval);
  }
   
  if (mul_sz) {
    pfile_log(pf, pfile_log_debug, "...multiply %u bits", 8 * mul_sz);
  }
  if (div_sz) {
    pfile_log(pf, pfile_log_debug, "...divide %u bits ", 8 * div_sz);
  }

  /* i either need to assume _accum1, _accum2, _sign, and _loop
     are all used (and come up with suitable sizes for _accum?,
     or generate the code twice -- once to get the sizes &
     once to actually generate the code.
     why is this? because the add code requires that both variables
     be in the same bank, so first generate the code, then allocate
     the variables, then generate the code again to make sure that
     any multi-byte adds work!! */
  lbl_tmp = pfile_lblptr_get(pf);
  /* define the processor (this needs to be changed) */
  for (pass = 0; pass < 2; pass++) {
    label_t    *lbl;

    pfile_log(pf, pfile_log_info, PIC_MSG_GENERATING_CODE, 1 + pass);
    if (1 == pass) {
      pic_code_reset(pf);
      pic_variable_alloc(pf, ct, vars);
      pfile_lblptr_set(pf, lbl_tmp);
    }

    pic_code_preamble(pf, ct, vars);
    for (code = pic_code_lst; code; code = code->link) {
      code->cmd = cmd;
    }
    for (cmd_ptr = cmd; cmd_ptr; cmd_ptr = cmd_link_get(cmd_ptr)) {
      if (cmd_flag_test(cmd_ptr, CMD_FLAG_REACHABLE)) {
        pic_cmd_out(pf, &cmd_ptr);
      }
    }
    /* once all of the commands are allocated, we'll do some analysis */
    /* first, on reset the following are cleared */
#if 1
    if (printval) {
      pic_printfn_create(pf);
    }
#endif

    lbl = pfile_label_find(pf, pfile_log_none, "_div");
    if (lbl) {
      pic_instr_append_label(pf, lbl);
      pic_divide_create(pf);
      label_release(lbl);
    }
    lbl = pfile_label_find(pf, pfile_log_none, "_sdiv");
    if (lbl) {
      pic_instr_append_label(pf, lbl);
      pic_sdivide_create(pf);
      label_release(lbl);
    }
    lbl = pfile_label_find(pf, pfile_log_none, "_mul");
    if (lbl) {
      pic_instr_append_label(pf, lbl);
      pic_multiply_create(pf);
      label_release(lbl);
    }
    if (1 == pass) {
      pic_array_setup_create(pf);
    }
    pic_instr_append(pf, pic_opcode_end);
  }
  /* now that the code is generated, reverse the list */
  {
    pic_code_t *ptr;
    pic_code_t *ptr_pv;
    pic_code_t *ptr_next;

    for (ptr_pv = 0, ptr = pic_code_lst; 
         ptr; 
         ptr_pv = ptr, ptr = ptr_next) {
      ptr_next = ptr->link;
      ptr->link = ptr_pv;
    }
    pic_code_lst = ptr_pv;
  }


  pfile_log(pf, pfile_log_info, PIC_MSG_FIXING_DATA_BITS);
#if 1
  {
    label_t       *lbl;
    pic_databits_t databits;

    /* first -- look for isr_preamble */
    lbl = pfile_label_find(pf, pfile_log_none, LABEL_PIC_ISR);
    if (lbl) {
      code = pic_code_label_find(pf, lbl);
      if (code) {
        databits.irp = pic_bitstate_clr;
        databits.rp1 = pic_bitstate_clr;
        databits.rp0 = pic_bitstate_clr;

        pic_code_databits_analyze(pf, code, &databits, 1);
      }
      label_release(lbl);
    }

    lbl = pfile_label_find(pf, pfile_log_err, LABEL_PIC_RESET);
    if (lbl) {
      code = pic_code_label_find(pf, lbl);
      if (code) {
        databits.irp = pic_bitstate_clr;
        databits.rp1 = pic_bitstate_clr;
        databits.rp0 = pic_bitstate_clr;

        pic_code_databits_analyze(pf, code, &databits, 1);
      }
      label_release(lbl);
    }
  }
#endif
  /* once the analysis is done, we need to run through & make
     sure no further label fixups are needed, or fix them up
     as necessary */
  ins_ct = pic_code_databits_post_analyze(pf);
  pfile_log(pf, pfile_log_debug, PIC_MSG_DBG_INSERTED_INST, ins_ct, 1);
  pcl = 0;
  for (code = pic_code_lst; code; code = code->link) {
    label_t *lbl;

    if (pic_opcode_org == code->op) {
      pcl = (size_t) value_const_get(code->literal);
    } 
    code->pc = pcl;
    lbl = code->lbl;
    if (lbl) {
      label_pc_set(lbl, pcl);
    }
    pcl++;
  }
#if 1
  /* if (pcl > 2048) */{
    /* need to analyze the code an insert pclath<4:3> maintainence
       as necessary */
    /* first -- look for isr_preamble */
    size_t ins;

    pass   = 0;
    ins_ct = 0;
    pfile_log(pf, pfile_log_info, PIC_MSG_FIXING_CODE_BITS);
    do {
      pic_branchbits_t branchbits;
      label_t         *lbl;

      for (code = pic_code_lst; code; code = code->link) {
        if (pic_opcode_org == code->op) {
          pcl = (size_t) value_const_get(code->literal);
        } 
        code->pc = pcl;
        lbl = code->lbl;
        if (lbl) {
          label_pc_set(lbl, pcl);
        }
#if 0
        code->branchbits.before.pclath4 = (pic_pcl < 4096)
          ? pic_bitstate_clr : pic_bitstate_unknown;
        code->branchbits.before.pclath3 = (pic_pcl < 2048)
          ? pic_bitstate_clr : pic_bitstate_unknown;
#else
        code->branchbits.before.pclath4 = pic_bitstate_unknown;
        code->branchbits.before.pclath3 = pic_bitstate_unknown;
#endif
        code->branchbits.action.pclath4 = pic_bitstate_unknown;
        code->branchbits.action.pclath3 = pic_bitstate_unknown;
        if (0xffff != pic_code_to_pcode(pf, code)) {
          pcl++;
        }
      }
      lbl = pfile_label_find(pf, pfile_log_none, LABEL_PIC_ISR);
      if (lbl) {
        code = pic_code_label_find(pf, lbl);
        if (code) {
          branchbits.pclath4 = pic_bitstate_clr;
          branchbits.pclath3 = pic_bitstate_clr;

          pic_code_branchbits_analyze(pf, code, &branchbits, 1);
        }
        label_release(lbl);
      }

      lbl = pfile_label_find(pf, pfile_log_err, LABEL_PIC_RESET);
      if (lbl) {
        code = pic_code_label_find(pf, lbl);
        if (code) {
          branchbits.pclath4 = pic_bitstate_clr;
          branchbits.pclath3 = pic_bitstate_clr;

          pic_code_branchbits_analyze(pf, code, &branchbits, 1);
        }
        label_release(lbl);
      }
#if 1
      ins = pic_branchbits_action_fixup(pf);
      ins_ct += ins;
#else
      ins = 0;
#endif
      pass++;
    } while (ins);
    pfile_log(pf, pfile_log_debug, PIC_MSG_DBG_INSERTED_INST, ins_ct, pass);
  }
#endif

  /* printf("PIC analyze max depth: %u\n", max_depth); */
  /* pic_cmd_dump(pf); */
  /* note: i'll also need to analyze starting at the interrupt vector (4) */
  pfile_write(pf, pfile_write_asm, "%32slist p=16f877, r=dec\n", "");
  /* set some useful macros */
  pfile_write(pf, pfile_write_asm, 
      "datahi_set macro\n"
      "  bsf _status, _rp1\n"
      "  endm\n");
  pfile_write(pf, pfile_write_asm,
      "datahi_clr macro\n"
      "  bcf _status, _rp1\n"
      "  endm\n");
  pfile_write(pf, pfile_write_asm,
      "datalo_set macro\n"
      "  bsf _status, _rp0\n"
      "  endm\n");
  pfile_write(pf, pfile_write_asm,
      "datalo_clr macro\n"
      "  bcf _status, _rp0\n"
      "  endm\n");
  pfile_write(pf, pfile_write_asm,
      "branchhi_set macro\n"
      "  bsf _pclath, 4\n"
      "  endm\n");
  pfile_write(pf, pfile_write_asm,
      "branchhi_clr macro\n"
      "  bcf _pclath, 4\n"
      "  endm\n");
  pfile_write(pf, pfile_write_asm,
      "branchlo_set macro\n"
      "  bsf _pclath, 3\n"
      "  endm\n");
  pfile_write(pf, pfile_write_asm,
      "branchlo_clr macro\n"
      "  bcf _pclath, 3\n"
      "  endm\n");

  {
    variable_t *vptr;

    /* dump the variables. note the _rp0, _rp1, _status and _pclath are used
     * in macros so they may not ever be marked as used. so i need to mark
     * them here */
    vptr = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_rp0");
    if (vptr) {
      variable_flag_set(vptr, VARIABLE_FLAG_USED);
      variable_release(vptr);
    }
    vptr = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_rp1");
    if (vptr) {
      variable_flag_set(vptr, VARIABLE_FLAG_USED);
      variable_release(vptr);
    }
    vptr = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_pclath");
    if (vptr) {
      variable_flag_set(vptr, VARIABLE_FLAG_USED);
      variable_release(vptr);
    }
    vptr = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_status");
    if (vptr) {
      variable_flag_set(vptr, VARIABLE_FLAG_USED);
      variable_release(vptr);
    }

    for (ii = 0; ii < ct; ii++) {

      for (vptr = vars[ii]; vptr; vptr = variable_link_get(vptr)) {
        variable_const_t vdef;

        if ((variable_flag_test(vptr, VARIABLE_FLAG_ASSIGNED)
              || variable_flag_test(vptr, VARIABLE_FLAG_CONST)
              || variable_flag_test(vptr, VARIABLE_FLAG_VOLATILE))
            && variable_flag_test(vptr, VARIABLE_FLAG_USED)) {
          if (variable_flag_test(vptr, VARIABLE_FLAG_CONST)) {
            if (variable_ct_get(vptr) == 1) {
              vdef = variable_const_get(vptr, 0);
              pfile_write(pf, pfile_write_asm, 
                  "%-31.31s EQU %u\n",
                  variable_name_get(vptr), vdef);
            }
          } else {
            vdef = variable_base_get(vptr);
            if (VARIABLE_BASE_UNKNOWN != vdef) {
              pfile_write(pf, pfile_write_asm, 
                  "%-31.31s EQU 0x%04x\n",
                  variable_name_get(vptr), vdef);
            }
          }
        }
      }
    }
  }
}

boolean_t pic_opcode_is_conditional(pic_opcode_t op)
{
  return (pic_opcode_incfsz == op)
    || (pic_opcode_decfsz == op)
    || (pic_opcode_btfsc == op)
    || (pic_opcode_btfss == op);
}

/* the register flags */
#define RFLAGS VARIABLE_FLAG_NOTEMP | VARIABLE_FLAG_VOLATILE

static const struct {
  const char     *name;
  variable_base_t base;
  variable_sz_t   sz;
  variable_ct_t   ct;
  flag_t          flags;
} pic_vars[] = {
  /* accumulators */
  {"_sign",       VARIABLE_BASE_UNKNOWN, 0, 1, VARIABLE_FLAG_VOLATILE},
  {"_accum1",     VARIABLE_BASE_UNKNOWN, 0, 1, VARIABLE_FLAG_VOLATILE},
  {"_accum2",     VARIABLE_BASE_UNKNOWN, 0, 1, VARIABLE_FLAG_VOLATILE},
  {"_accum3",     VARIABLE_BASE_UNKNOWN, 0, 1, VARIABLE_FLAG_VOLATILE},
  {"_accum4",     VARIABLE_BASE_UNKNOWN, 0, 1, VARIABLE_FLAG_VOLATILE},
  {"_delay0",     VARIABLE_BASE_UNKNOWN, 0, 1, VARIABLE_FLAG_VOLATILE},
  {"_delay1",     VARIABLE_BASE_UNKNOWN, 0, 1, VARIABLE_FLAG_VOLATILE},
  {"_delay2",     VARIABLE_BASE_UNKNOWN, 0, 1, VARIABLE_FLAG_VOLATILE},
  /* state saving bits */
  {"_isrcodegen", VARIABLE_BASE_UNKNOWN, 0, 1, VARIABLE_FLAG_VOLATILE},
  {"_codegen",    VARIABLE_BASE_UNKNOWN, 0, 1, VARIABLE_FLAG_VOLATILE},
};

void pic_init(pfile_t *pf)
{
  size_t ct;

  for (ct = 0; ct < COUNT(pic_vars); ct++) {
    variable_t *var;

    if (result_ok == pfile_variable_alloc(pf, pic_vars[ct].name,
          pic_vars[ct].flags,
        pic_vars[ct].sz, pic_vars[ct].ct, 0, &var)) {
      variable_base_set(var, pic_vars[ct].base);
      variable_release(var);
    }
  }
}

