/************************************************************
 **
 ** pic_inst.c : PIC instruction generation definitions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <assert.h>

#include "pic.h"
#include "pic_var.h"
#include "pic_inst.h"

void pic_instr_append_label(pfile_t *pf, label_t *lbl)
{
  pic_code_get(pf, lbl, pic_opcode_none);
}

void pic_instr_append_f_d(pfile_t *pf, pic_opcode_t op, value_t *f, 
    size_t ofs, pic_opdst_t dst)
{
  pic_code_t *code;

  code = pic_code_get(pf, 0, op);
  if (code) {
    if (f) {
      value_lock(f);
      value_flag_set(f, VARIABLE_FLAG_USED);
    }
    code->val = f;
    code->ofs = ofs;
    code->dst = dst;
  }
}

void pic_instr_append_reg_d(pfile_t *pf, pic_opcode_t op, const char *reg,
  pic_opdst_t dst)
{
  value_t *src;

  src = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF, reg);
  if (src) {
    pic_instr_append_f_d(pf, op, src, 0, dst);
    value_release(src);
  }
}


void pic_instr_append_f(pfile_t *pf, pic_opcode_t op, value_t *f, 
    size_t ofs)
{
  pic_code_t *code;

  code = pic_code_get(pf, 0, op);
  if (code) {
    if (f) {
      value_flag_set(f, VARIABLE_FLAG_USED);
      value_lock(f);
    }
    code->val = f;
    code->ofs = ofs;
  }
}

void pic_instr_append_reg(pfile_t *pf, pic_opcode_t op, const char *reg)
{
  value_t *tmp;

  tmp = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF, reg);
  if (tmp) {
    pic_instr_append_f(pf, op, tmp, 0);
    value_release(tmp);
  }
}


void pic_instr_append(pfile_t *pf, pic_opcode_t op)
{
  pic_code_get(pf, 0, op);
}

void pic_instr_append_f_b(pfile_t *pf, pic_opcode_t op, value_t *f, 
    size_t ofs, value_t *bit)
{
  pic_code_t *code;

  code = pic_code_get(pf, 0, op);
  if (code) {
    if (f) {
      value_flag_set(f, VARIABLE_FLAG_USED);
      value_lock(f);
    }
    code->val = f;
    code->ofs = ofs;
    if (bit) {
      value_flag_set(bit, VARIABLE_FLAG_USED);
      value_lock(bit);
    }
    code->literal = bit;
  }
}

void pic_instr_append_reg_bn(pfile_t *pf, pic_opcode_t op, 
    const char *reg, variable_const_t bit)
{
  value_t *vreg;

  vreg = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF, reg);
  if (vreg) {
    value_t *vflag;

    if (result_ok == pfile_constant_get(pf, bit, &vflag)) {
      pic_instr_append_f_b(pf, op, vreg, 0, vflag);
      value_release(vflag);
    }
    value_release(vreg);
  }
}

void pic_instr_append_reg_flag(pfile_t *pf, pic_opcode_t op, 
    const char *reg, const char *flag)
{
  value_t *vreg;

  vreg = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF, reg);
  if (vreg) {
    value_t *vflag;

    vflag = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF, flag);
    if (vflag) {
      pic_instr_append_f_b(pf, op, vreg, 0, vflag);
      value_release(vflag);
    }
    value_release(vreg);
  }
}

void pic_instr_append_f_bn(pfile_t *pf, pic_opcode_t op,
    value_t *f, size_t ofs, variable_const_t n)
{
  value_t *vn;

  if (result_ok == pfile_constant_get(pf, n, &vn)) {
    pic_instr_append_f_b(pf, op, f, ofs + value_baseofs_get(f), vn);
    value_release(vn);
  }
}

void pic_instr_append_w_k(pfile_t *pf, pic_opcode_t op, 
    value_t *k)
{
  pic_code_t *code;

  code = pic_code_get(pf, 0, op);
  if (code) {
    if (k) {
      value_lock(k);
      value_flag_set(k, VARIABLE_FLAG_USED);
    }
    code->literal = k;
  }
}

void pic_instr_append_w_kn(pfile_t *pf, pic_opcode_t op,
    variable_const_t n)
{
  value_t *vn;

  if (result_ok == pfile_constant_get(pf, n, &vn)) {
    pic_instr_append_w_k(pf, op, vn);
    value_release(vn);
  }
}


void pic_instr_append_n(pfile_t *pf, pic_opcode_t op, label_t *dst)
{
  pic_code_t *code;

  code = pic_code_get(pf, 0, op);
  if (code) {
    if (dst) {
      label_lock(dst);
    }
    code->brdst = dst;
  }
}

unsigned pic_code_to_pcode(pfile_t *pf, const pic_code_t *code)
{
  unsigned pcode;
  unsigned literal;
  unsigned val;
  unsigned dst;
  unsigned bit;
  unsigned lit8;
  unsigned lit10; /* branch literal */

  if (code->literal) {
    if (value_flag_test(code->literal, VARIABLE_FLAG_CONST)) {
      literal = value_const_get(code->literal);
    } else {
      literal = value_base_get(code->literal)
        + value_baseofs_get(code->literal);
    }
  } else {
    literal = 0xffff;
  }
  lit10   = (code->brdst) ? label_pc_get(code->brdst) : 0;

  if (code->val) {
    if ((value_is_array(code->val) && (value_subscript_get(code->val)))
      || (VALUE_SUBSCRIPT_DUMMY == value_subscript_get(code->val))) {
      val = 0; /* _ind */
    } else {
      val = (value_base_get(code->val)
          + pic_var_ofs_get(pf, code->val, code->ofs))
        & 0x7f;
    }
  } else {
    val = 0xffff;
  }
  bit     = (literal & 0x07);
  lit8    = literal & 0x00ff; /* 8 bit literal */
  dst     = (pic_opdst_f == code->dst) << 7;
  pcode   = 0xffff; /* an impossible code (it's greater than 14 bits) */

  if (((pic_opcode_movlw == code->op) && code->brdst)) {
    /* this is a special instruction --> load w with the high 5 bits of brdst */
    lit8 = lit10 >> 8;
  }

  switch (code->op) {
    /* directives (no code) */
    case pic_opcode_org:    /* ORG xxxx */
    case pic_opcode_end:    /* END      */
    case pic_opcode_none:   /* no code here */
      break;
    case pic_opcode_nop:    pcode = 0x0000; break;
    /* file operator (code, dst, val) */
    case pic_opcode_movwf:  pcode = 0x0000 | 0x80 | val; break;
    case pic_opcode_clrf:   pcode = 0x0100 | 0x80 | val; break;
    case pic_opcode_subwf:  pcode = 0x0200 |  dst | val; break;
    case pic_opcode_decf:   pcode = 0x0300 |  dst | val; break;
    case pic_opcode_iorwf:  pcode = 0x0400 |  dst | val; break;
    case pic_opcode_andwf:  pcode = 0x0500 |  dst | val; break;
    case pic_opcode_xorwf:  pcode = 0x0600 |  dst | val; break;
    case pic_opcode_addwf:  pcode = 0x0700 |  dst | val; break;
    case pic_opcode_movf:   pcode = 0x0800 |  dst | val; break;
    case pic_opcode_comf:   pcode = 0x0900 |  dst | val; break;
    case pic_opcode_incf:   pcode = 0x0a00 |  dst | val; break;
    case pic_opcode_decfsz: pcode = 0x0b00 |  dst | val; break;
    case pic_opcode_rrf:    pcode = 0x0c00 |  dst | val; break;
    case pic_opcode_rlf:    pcode = 0x0d00 |  dst | val; break;
    case pic_opcode_swapf:  pcode = 0x0e00 |  dst | val; break;
    case pic_opcode_incfsz: pcode = 0x0f00 |  dst | val; break;
    /* simple opcodes (no operands) */
    case pic_opcode_retfie: pcode = 0x0009; break;
    case pic_opcode_ret:    pcode = 0x0008; break;
    case pic_opcode_sleep:  pcode = 0x0063; break;
    case pic_opcode_clrwdt: pcode = 0x0064; break;
    /* the following doesn't look right to me based on the microchip doc's, but
     * it is what MPLAB puts out */
    case pic_opcode_clrw:   pcode = 0x0103; break;
    /* bit oriented (3 bit literal) */
    case pic_opcode_bcf:    pcode = 0x1000 | (bit << 7) | val; break;
    case pic_opcode_bsf:    pcode = 0x1400 | (bit << 7) | val; break;
    case pic_opcode_btfsc:  pcode = 0x1800 | (bit << 7) | val; break;
    case pic_opcode_btfss:  pcode = 0x1c00 | (bit << 7) | val; break;
    /* literal & w  (8 bit literal) */ 
    case pic_opcode_movlw:  pcode = 0x3000 | lit8; break;
    case pic_opcode_retlw:  pcode = 0x3400 | lit8; break;
    case pic_opcode_iorlw:  pcode = 0x3800 | lit8; break;
    case pic_opcode_andlw:  pcode = 0x3900 | lit8; break;
    case pic_opcode_xorlw:  pcode = 0x3a00 | lit8; break;
    case pic_opcode_sublw:  pcode = 0x3c00 | lit8; break;
    case pic_opcode_addlw:  pcode = 0x3e00 | lit8; break;
    /* branching (10 bit literal) */
    case pic_opcode_call:   pcode = 0x2000 | lit10; break;
    case pic_opcode_goto:   pcode = 0x2800 | lit10; break;
    /* special (pseudo) opcodes */
    case pic_opcode_datalo_set: /* bsf _status: _rp0 */
      pcode = 0x1400 | (5 << 7) | 0x0003;
      break;
    case pic_opcode_datalo_clr: /* bcf _status: _rp0 */
      pcode = 0x1000 | (5 << 7) | 0x0003;
      break;
    case pic_opcode_datahi_set: /* bsf _status: _rp1 */
      pcode = 0x1400 | (6 << 7) | 0x0003;
      break;
    case pic_opcode_datahi_clr: /* bcf _status: _rp1 */
      pcode = 0x1000 | (6 << 7) | 0x0003;
      break;
    case pic_opcode_branchlo_set: /* bsf _pclath: 3 */
      pcode = 0x1400 | (3 << 7) | 0x000a;
      break;
    case pic_opcode_branchlo_clr: /* bcf _pclath: 3 */
      pcode = 0x1000 | (3 << 7) | 0x000a;
      break;
    case pic_opcode_branchhi_set: /* bsf _pclath: 4 */
      pcode = 0x1400 | (4 << 7) | 0x000a;
      break;
    case pic_opcode_branchhi_clr: /* bcf _pclath: 4 */
      pcode = 0x1000 | (4 << 7) | 0x000a;
      break;
  }
  return pcode;
}


