/************************************************************
 **
 ** pic_opfn.c : PIC built-in function generation
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include "pfile.h"
#include "bsc_file.h"
#include "bsc_intr.h"
#include "pic.h"
#include "pic_op.h"
#include "pic_opfn.h"

/*
 * NAME
 *   pic_multiply_create
 *
 * DESCRIPTION
 *   create an n-bit multiply
 *
 * PARAMETERS
 *   pf : pfile
 *
 * RETURN
 *
 * NOTES
 *   accum1 = val 1
 *   accum2 = val 2
 *   algorithm
 *      for ii = 1 to bits
 *         accum3 <<= 1
 *         accum1 <<= 1
 *         if (no carry)
 *           accum3 += accum2
 *         end if
 *      next
 *   on exit, accum3 holds the result
 */
void pic_multiply_create(pfile_t *pf)
{
  label_t *top;
  label_t *skip;
  value_t *accum1;
  value_t *accum2;
  value_t *accum3;
  value_t *c_1;
  value_t *loop;

  accum1 = pic_accum_get(pf, 1, mul_sz, 0);
  accum2 = pic_accum_get(pf, 2, mul_sz, 0);
  accum3 = pic_accum_get(pf, 3, mul_sz, 0);
  c_1 = 0;
  pfile_constant_get(pf, 1, &c_1);
  loop   = pic_loop_get(pf);
  top = 0;
  pfile_label_alloc(pf, 0, 0, &top);
  skip = 0;
  pfile_label_alloc(pf, 0, 0, &skip);

  if (accum1 && accum2 && accum3 && c_1 && loop && top && skip) {
    pic_instr_append_w_kn(pf, pic_opcode_movlw, 8 * mul_sz);
    pic_instr_append_f(pf, pic_opcode_movwf, loop, 0);
    pic_instr_append_label(pf, top);
    pic_op(pf, operator_shift_left, accum3, c_1, 0);
    pic_op(pf, operator_shift_left, accum1, c_1, 0);
    pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_c");
    pic_instr_append_n(pf, pic_opcode_goto, skip);
    pic_op(pf, operator_add, accum3, accum2, 0);
    pic_instr_append_label(pf, skip);
    pic_instr_append_f_d(pf, pic_opcode_decfsz, loop, 0, pic_opdst_f);
    pic_instr_append_n(pf, pic_opcode_goto, top);
    pic_instr_append(pf, pic_opcode_ret);
  }
  label_release(skip);
  label_release(top);
  value_release(loop);
  value_release(c_1);
  value_release(accum3);
  value_release(accum2);
  value_release(accum1);
}

/*
 * NAME
 *    pic_divide_create
 *
 * DESCRIPTION
 *    create an n-bit divide
 *
 * PARAMETERS
 *    pf : pfile
 *
 * RETURN
 *
 * NOTES
 *   entry:
 *     accum1 = dividend
 *     accum4 = divisor
 *   algorithm:
 *     for ii = 1 to bits
 *       accum3 <<= 1
 *       accum1 <<= 1
 *       accum2 = accum2 << 1 | carry from prev. op
 *       if (accum2 >= accum4)
 *         accum2 -= accum4
 *         accum3 |= 1
 *       end if
 *     next 
 */
void pic_divide_create(pfile_t *pf)
{
  value_t *accum1;
  value_t *accum2;
  value_t *accum3;
  value_t *accum4;
  value_t *loop;
  value_t *c_1;
  value_t *c_0;
  label_t *top;
  label_t *skip;

  accum1 = pic_accum_get(pf, 1, div_sz, 0);
  accum2 = pic_accum_get(pf, 2, div_sz, 0);
  accum3 = pic_accum_get(pf, 3, div_sz, 0);
  accum4 = pic_accum_get(pf, 4, div_sz, 0);
  c_1 = 0;
  pfile_constant_get(pf, 1, &c_1);
  c_0 = 0;
  pfile_constant_get(pf, 0, &c_0);

  loop   = pic_loop_get(pf);
  top = 0;
  pfile_label_alloc(pf, 0, 0, &top);
  skip = 0;
  pfile_label_alloc(pf, 0, 0, &skip);

  if (accum1 && accum2 && accum3 && accum4 && c_1 
      && c_0 && loop && top && skip) {
    /* # bits --> loop */
    pic_instr_append_w_kn(pf, pic_opcode_movlw, 8 * div_sz);
    pic_instr_append_f(pf, pic_opcode_movwf, loop, 0);
    pic_op(pf, operator_assign, accum2, c_0, 0);
    pic_instr_append_label(pf, top);
    /* accum3 <<= 1 */
    pic_op(pf, operator_shift_left, accum3, c_1, 0);
    /* {accum1:accum2} <<= 1 */
    /* note : setup accum1 as double size, so the rotate also goes
              through accum2 */
    value_sz_set(accum1, value_sz_get(accum1) * 2);
    pic_op(pf, operator_shift_left, accum1, c_1, 0);
    /* accum2 -= accum4 */
    pic_op(pf, operator_sub, accum2, accum4, 0);
    /* if accum4 > accum2, add it back */
    pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_c");
    pic_instr_append_n(pf, pic_opcode_goto, skip);
    pic_op(pf, operator_add, accum2, accum4, 0);
    /* status<c> is guarenteed to be set here, so this simply skips the next
      instruction */
    pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_c");
    pic_instr_append_label(pf, skip);
    pic_instr_append_f_bn(pf, pic_opcode_bsf, accum3, 0, 0);
    pic_instr_append_f_d(pf, pic_opcode_decfsz, loop, 0, pic_opdst_f);
    pic_instr_append_n(pf, pic_opcode_goto, top);
    pic_instr_append(pf, pic_opcode_ret);
  }
  label_release(skip);
  label_release(top);
  value_release(loop);
  value_release(c_1);
  value_release(c_0);
  value_release(accum4);
  value_release(accum3);
  value_release(accum2);
  value_release(accum1);
}

/*
 * NAME
 *    pic_sdivide_create
 *
 * DESCRIPTION
 *    create a signed divide
 *
 * PARAMETERS
 *    pf
 *
 * RETURN
 *
 * NOTES
 *   on entry:
 *   algorithm:
 *     if accum1 & accum4 are not the same sign, the result is (-)
 *     accum1 = abs(accum1)
 *     accum4 = abs(accum4)
 *     call divide (above)
 *     if (-) then accum3 = -accum3
 */
void pic_sdivide_create(pfile_t *pf)
{
  value_t *accum1;
  value_t *accum4;
  value_t *accum3;
  value_t *sign;
  label_t *lbl;

  accum1 = pic_accum_get(pf, 1, div_sz, 0);
  accum4 = pic_accum_get(pf, 4, div_sz, 0);
  accum3 = pic_accum_get(pf, 3, div_sz, 0);
  sign = pic_sign_get(pf);
  lbl = 0;
  pfile_label_alloc(pf, 0, 0, &lbl);

  if (accum1 && accum4 && accum3 && sign && lbl) {
    pic_instr_append(pf, pic_opcode_clrw);
    pic_instr_append_f_bn(pf, pic_opcode_btfss, accum1, 3, 7);
    pic_instr_append_n(pf, pic_opcode_goto, lbl);
    /* accum1 = -accum1 */
    pic_op(pf, operator_cmpb, accum1, accum1, 0);
    pic_op(pf, operator_incr, accum1, accum1, 0);
    pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
    pic_instr_append_label(pf, lbl);
    label_release(lbl);
    lbl = 0;
    /* sign = 0 (unsigned) 1 (signed) */
    pic_instr_append_f(pf, pic_opcode_movwf, sign, 0);

    if (result_ok == pfile_label_alloc(pf, 0, 0, &lbl)) {
      pic_instr_append(pf, pic_opcode_clrw);
      pic_instr_append_f_bn(pf, pic_opcode_btfss, accum4, 3, 7);
      pic_instr_append_n(pf, pic_opcode_goto, lbl);
      /* accum4 = -accum4 */
      pic_op(pf, operator_cmpb, accum4, 0, 0);
      pic_op(pf, operator_incr, accum4, 0, 0);
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
      pic_instr_append_label(pf, lbl);
      /* sign = 0 (unsigned) 1 (signed) */
      pic_instr_append_f_d(pf, pic_opcode_xorwf, sign, 0, pic_opdst_f);

      label_release(lbl);
      lbl = pfile_label_find(pf, pfile_log_none, "_div");
      if (lbl) {
        pic_instr_append_n(pf, pic_opcode_call, lbl);
        label_release(lbl);
        /* what to do with the result */
        pic_instr_append_f_d(pf, pic_opcode_movf, sign, 0, pic_opdst_w);
        pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
        pic_instr_append(pf, pic_opcode_ret);
        /* invert the result */
        pic_op(pf, operator_cmpb, accum3, 0, 0);
        pic_op(pf, operator_incr, accum3, 0, 0);
        pic_instr_append(pf, pic_opcode_ret);
      }
    }
  }
  label_release(lbl);
  value_release(sign);
  value_release(accum3);
  value_release(accum4);
  value_release(accum1);
}

/*
 * NAME
 *   pic_array_setup_create
 *
 * DESCRIPTION
 *   setup an array that spans banks
 *
 * PARAMETERS
 *   pf
 *
 * RETURN
 *
 * NOTES
 *   on entry, accum4 holds the array index, accum3 holds the
 *     highest location used in a bank and IRP is set correctly for
 *     array index 0
 *   on exit, _fsr and <status:_irp>
 *     are set correctly for array access
 *   the following naming convention is used:
 *     _array_setup_[sz]_[bank]
 *   because different values are used
 *   currently, this only 
 */
void pic_array_setup_label_create(char *dst, size_t sz, size_t bank)
{
  sprintf(dst, "_pic_asetup_%u_%u", (unsigned) sz, (unsigned) bank);
}

void pic_array_setup_create(pfile_t *pf)
{
  size_t                 sz;
  value_t               *accum3;
  value_t               *accum4;
  value_t               *sign;
  value_t               *fsr;

  accum3 = pic_accum_get(pf, 3, 2, 0); /* highest accessible in bank */
  accum4 = pic_accum_get(pf, 4, 2, 0); /* index we're looking for    */
  sign   = pic_sign_get(pf);           /* scratch                    */
  fsr    = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_fsr");

  for (sz = 1; sz <= 4; sz <<= 1) {
    const pic_bank_info_t *bank;
    unsigned               bank_no;
    boolean_t              found;
    label_t               *done;
    label_t               *done_irp;

    found = boolean_false;

    done_irp = 0;
    done     = 0;

    for (bank_no = 0, bank = pic_blist;
         bank && bank->link;
         bank_no++, bank = bank->link) {
      label_t *label;
      char     label_name[32];
      size_t   highest_used;

      highest_used = ((bank->hi + 1 - bank->lo) / sz ) * sz + ( sz - 1);
      pic_array_setup_label_create(label_name, sz, bank_no);
      label = pfile_label_find(pf, pfile_log_none, label_name);
      if (label) {
        if (!found) {
          found = boolean_true;
          pfile_label_alloc(pf, 0, 0, &done);
          pfile_label_alloc(pf, 0, 0, &done_irp);
        }
      } 
      if (found) {
        value_t *c_hi; /* next bank's highest location */
        value_t *c_lo; /* next bank's lowest location */
        if (label) {
          pic_instr_append_label(pf, label);
          label_release(label);
        }

        c_hi = 0;
        c_lo = 0;
        if ((result_ok == pfile_constant_get(pf, bank->link->hi + 1, &c_hi)) 
          && (result_ok == pfile_constant_get(pf, bank->link->lo, &c_lo))) {
          /* sign = (accum4 < accum3) */
          pic_op(pf, operator_lt, sign, accum4, accum3);
          pic_instr_append_f_bn(pf, pic_opcode_btfsc, sign, 0, 0);
          if (bank->lo < 256) {
            pic_instr_append_n(pf, pic_opcode_goto, done);
          } else {
            pic_instr_append_n(pf, pic_opcode_goto, done_irp);
          }
          /* accum4 += c_lo */
          /* accum4 -= accum3 (highest location in current bank) */
          pic_op(pf, operator_add, accum4, c_lo, 0);
          pic_op(pf, operator_sub, accum4, accum3, 0);
          /* accum3 = c_hi */
          pic_op(pf, operator_assign, accum3, c_hi, 0);
        }
        value_release(c_hi);
        value_release(c_lo);
      }
    }
    if (found) {
      if (done_irp) {
        pic_instr_append_label(pf, done_irp);
        pic_instr_append_reg_flag(pf, pic_opcode_bsf, "_status", "_irp");
        label_release(done_irp);
      }
      pic_instr_append_label(pf, done);
      label_release(done);
      /* _fsr = _accum4 */
      pic_op(pf, operator_assign, fsr, accum4, 0);
      pic_instr_append(pf, pic_opcode_ret);
    }
  }
  value_release(fsr);
  value_release(sign);
  value_release(accum4);
  value_release(accum3);
}

