/************************************************************
 **
 ** pic_var.c : PIC variable allocation definitions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include "pic.h"
#include "pic_var.h"

/*
 * NAME
 *   pic_variable_alloc
 *
 * DESCRIPTION
 *   allocate space for all variables
 *
 * PARAMETERS
 *   pf
 *
 * RETURN
 *
 * NOTES
 *   this is dummied up for now. eventually i'll need to sort
 *   the variables by size for a more optimimal filling
 */
/* this is the union of  blist with {lo,hi} */
void pic_blist_add(pfile_t *pf, pic_bank_info_t **pblist, 
    size_t lo, size_t hi)
{
  pic_bank_info_t *ptr;
  pic_bank_info_t *pv;
  pic_bank_info_t *blist;

  blist = *pblist;
  for (ptr = blist, pv = 0; 
       ptr && (lo > ptr->hi);
       pv = ptr, ptr = ptr->link)
    ;
  /* now, ptr is the best guess! */
  if (!ptr || (hi < ptr->lo)) {
    /* create a new entry & insert it after pv */
    ptr = MALLOC(sizeof(*ptr));
    if (!ptr) {
      pfile_log_syserr(pf, result_memory);
    } else {
      ptr->link = (pv) ? pv->link : 0;
      if (pv) {
        pv->link = ptr;
      } else {
        blist = ptr;
      }
      ptr->lo = lo;
      ptr->hi = hi;
      ptr->next = ptr->lo;
    }
  } else {
    /* here we know:
          ptr exists
          lo <  ptr->hi
          hi >= ptr->lo */
    if (lo < ptr->lo) {
      /* expand down */
      ptr->lo = lo;
      ptr->next = ptr->lo;
    }
    if (hi > ptr->hi) {
      /* expand up */
      ptr->hi = hi;
      /* look to see if this add absorbes the next element */
      pv = ptr;
      ptr = ptr->link;
      while (ptr && (pv->hi >= ptr->lo)) {
        if (pv->hi < ptr->hi) {
          pv->hi = ptr->hi;
        }
        pv->link = ptr->link;
        FREE(ptr);
        ptr = pv->link;
      }
    }
  }
  *pblist = blist;
}

/* remove {base,sz} from the blist */
static void pic_blist_remove(pfile_t *pf, pic_bank_info_t **pblist,
    size_t base, size_t sz)
{
  pic_bank_info_t *ptr;
  pic_bank_info_t *pv;
  pic_bank_info_t *blist;
  size_t           lo;
  size_t           hi;

  blist = *pblist;
  lo = base;
  hi = base + sz - 1;

  for (ptr = blist, pv = 0;
       ptr && (lo > ptr->hi);
       pv = ptr, ptr = ptr->link)
    ;
  if (ptr && (lo <= ptr->hi) && (hi >= ptr->lo)) {
    /* lo <= ptr->hi
       ptr->lo <= hi <= ptr->hi */
    if (lo <= ptr->lo) {
      ptr->lo = hi + 1;
      ptr->next = ptr->lo;
    } 
    if (hi >= ptr->hi) {
      ptr->hi = lo - 1;
    }
    /* if (ptr->hi < ptr->lo), this bit is gone */
    if (ptr->hi < ptr->lo) {
      if (pv) {
        pv->link = ptr->link;
      } else {
        *pblist = ptr->link;
      }
      FREE(ptr);
    } else if ((ptr->lo < lo) && (hi < ptr->hi)) {
      /* split this into:
           {ptr->lo, lo - 1}
           {hi + 1, ptr->hi} */
      pv = ptr;
      ptr = MALLOC(sizeof(*ptr));
      if (!ptr) {
        pfile_log_syserr(pf, result_memory);
      } else {
        ptr->link = pv->link;
        pv->link  = ptr;
        ptr->hi   = pv->hi;
        ptr->lo   = hi + 1;
        ptr->next = ptr->lo;
        pv->hi    = lo - 1;
      }
    }
  }
  *pblist = blist;
}

static void pic_blist_dump(pfile_t *pf, pic_bank_info_t *blist)
{
  const pic_bank_info_t *bptr;

  for (bptr = blist; bptr; bptr = bptr->link) {
    pfile_log(pf, pfile_log_debug, 
        PIC_MSG_DBG_BANK, bptr->lo, bptr->hi);
  }
}

static void codegen_base_set(variable_t *var, variable_base_t *base)
{
  if (variable_sz_get(var)) {
    variable_base_set(var, *base);
    *base += variable_sz_get(var);
  }
}

static int pic_variable_base_set(pfile_t *pf, pic_bank_info_t *blist,
    variable_t *vptr)
{
  size_t           need;
  variable_sz_t    vptr_sz;
  variable_ct_t    vptr_ct;
  pic_bank_info_t *bank;

  vptr_sz = variable_sz_get(vptr);
  vptr_ct = variable_ct_get(vptr);
  need = vptr_sz * vptr_ct;
  if (need) {
    /* first we'll see if the entire thing can fit in a bank.
       If so, good! */
    for (bank = blist; 
         bank && (bank->next + need > bank->hi); 
         bank = bank->link)
      ;
    if (bank) {
      variable_base_set(vptr, bank->next);
      bank->next += need;
    } else {
      /* this cannot fit into a single bank. If it's an array,
         let it span multiple banks but remember that
         (1) an entire element must fit in a single back
         (2) the banks must be contiguous (the array
             elements must begin at the beginning of each bank
             except the first one */
      variable_ct_t have;
      size_t        bank_ct;

      bank = blist;
      bank_ct = 0;
      do {
        have = (bank->hi - bank->next) / variable_sz_get(vptr);
        if (have) {
          pic_bank_info_t *bptr;

          bank_ct = 1;
          for (bptr = bank->link;
               (have < vptr_ct) 
               && bptr 
               && (bptr->next == bptr->lo);
               bptr = bptr->link) {
            have += (bptr->hi + 1 - bptr->next) / variable_sz_get(vptr);
            bank_ct++;
          }
        }
        if (have < vptr_ct) {
          bank = bank->link;
        }
      } while (bank && (have < vptr_ct));
      /* if have < vptr_ct then we've no room for this */
      if (have < variable_ct_get(vptr)) {
        pfile_log(pf, pfile_log_err, PIC_MSG_NO_DATA_SPACE,
          variable_name_get(vptr));
      } else {
        pfile_log(pf, pfile_log_warn, PIC_MSG_SPANS_BANKS,
            variable_name_get(vptr), bank_ct);
        variable_base_set(vptr, bank->next);
        have = 0;
        do {
          variable_ct_t bhave;

          /* bhave = max # of element that will fit in this bank */
          bhave = (bank->hi + 1 - bank->next) / vptr_sz;
          if (have + bhave > vptr_ct) {
            bhave = vptr_ct - have;
          }
          bank->next += bhave * vptr_sz;
          have += bhave;
          bank = bank->link;
        } while (have < vptr_ct);
      }
    }
    /* total += need; */
  }
  return 0;
}

/* create the code generator variable block */
static void pic_codegen_block_create(pfile_t *pf, pic_bank_info_t *blist)
{
  variable_t      *codegen;
  variable_t      *isrcodegen;
  variable_t      *accum1;
  variable_t      *accum2;
  variable_t      *accum3;
  variable_t      *accum4;
  variable_t      *delay0;
  variable_t      *delay1;
  variable_t      *delay2;
  variable_t      *sign;
  variable_t      *loop;
  variable_sz_t    need_accum;
  variable_sz_t    need_delay;
  variable_sz_t    need_misc;
  variable_sz_t    need;

  codegen = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, 
      "_codegen");
  isrcodegen = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, 
      "_isrcodegen");
  accum1 = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_accum1");
  accum2 = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_accum2");
  accum3 = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_accum3");
  accum4 = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_accum4");
  delay0 = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_delay0");
  delay1 = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_delay1");
  delay2 = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_delay2");
  sign = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_sign");
  loop = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_loop");
  variable_sz_set(sign, 1);
  need_accum = variable_sz_get(accum1)
    + variable_sz_get(accum2)
    + variable_sz_get(accum3)
    + variable_sz_get(accum4);
  need_delay = + variable_sz_get(delay0)
    + variable_sz_get(delay1)
    + variable_sz_get(delay2);
  need_misc = + variable_sz_get(sign)
    + variable_sz_get(loop);
  need = need_accum;
  if (need_delay > need) {
    need = need_delay;
  }
  need += need_misc;
  if (need) {
    variable_base_t base;
    label_t        *lbl;

    variable_sz_set(codegen, need);
    variable_sz_set(isrcodegen, need);
    /* codegen always starts at the beginning of the first bank,
     * and _loop must always be the first element (otherwise clear
     * will break horribly) */
    base = variable_base_get(loop);
    if (base + need > pic_blist->hi) {
      pfile_log(pf, pfile_log_err, 
          PIC_MSG_CANNOT_ALLOCATE_CODEGEN,
          need);
    } else {
      variable_base_set(codegen, base);
      pic_blist_remove(pf, &pic_blist, base, need);
    }
    lbl = pfile_isr_entry_get(pf);
    if (lbl) {
      pic_variable_base_set(pf, blist, isrcodegen);
      label_release(lbl);
    }
    codegen_base_set(loop, &base);
    codegen_base_set(sign, &base);
    codegen_base_set(accum1, &base);
    codegen_base_set(accum2, &base);
    codegen_base_set(accum3, &base);
    codegen_base_set(accum4, &base);
    base = variable_base_get(codegen);
    codegen_base_set(delay0, &base);
    codegen_base_set(delay1, &base);
    codegen_base_set(delay2, &base);
  }
  variable_release(loop);
  variable_release(sign);
  variable_release(delay2);
  variable_release(delay1);
  variable_release(delay0);
  variable_release(accum4);
  variable_release(accum3);
  variable_release(accum2);
  variable_release(accum1);
  variable_release(isrcodegen);
  variable_release(codegen);
}

/* find all variables that the user has placed & make sure these
   positions do not get reused */
pic_bank_info_t *pic_blist;
void pic_variable_alloc(pfile_t *pf, size_t ct, variable_t **var)
{
  size_t           ii;
  size_t           total;
  variable_t      *vptr;


  pfile_log(pf, pfile_log_info, PIC_MSG_ALLOCATING_VARS);
  /* remove any preset variables */
  for (ii = 0; ii < ct; ii++) {
    for (vptr = var[ii]; vptr; vptr = variable_link_get(vptr)) {
      if (VARIABLE_BASE_UNKNOWN != variable_base_get(vptr)) {
        variable_base_t base;

        base = variable_base_get(vptr);
        variable_region_set(vptr, base >> 7); /* at a minimum, it must exist here */
        /*if (0 != variable_base_get(vptr)) */{
          variable_region_t jj;

          for (jj = 0; jj < 8; jj++) {
            /* remove the variable from *each* bank in which it exists,
             * and make sure it's base is in the lowest bank in which
             * is exists */
            if (variable_region_test(vptr, jj)) {
              variable_base_t rbase;

              rbase = (jj << 7) | (base & 0x7f);
              if (rbase < base) {
                base = rbase;
              }
              pic_blist_remove(pf, &pic_blist, variable_base_get(vptr),
                    variable_sz_get(vptr) * variable_ct_get(vptr));
            }
          }
          variable_base_set(vptr, base);
        }
      }
    }
  }
  /* setup the next elements of blist
     note: i keep the next element in case i decide to retry the
           variable allocation using a different method */
  pic_blist_dump(pf, pic_blist);
  /* finally, allocate some variables! */
  pic_codegen_block_create(pf, pic_blist);
  total = 0;
  for (ii = 0; ii < ct; ii++) {
    for (vptr = var[ii]; vptr; vptr = variable_link_get(vptr)) {
      if ((VARIABLE_BASE_UNKNOWN == variable_base_get(vptr)) 
        && !variable_flag_test(vptr, VARIABLE_FLAG_CONST)
        && variable_flag_test(vptr, VARIABLE_FLAG_ASSIGNED)
        && variable_flag_test(vptr, VARIABLE_FLAG_USED)) {
        /* find the first bank that will hold this variable */
        pic_variable_base_set(pf, pic_blist, vptr);
        total += variable_sz_get(vptr);
      }
    }
  }
  pfile_log(pf, pfile_log_debug, PIC_MSG_DBG_DATA_USED, total);
}

void pic_blist_info_log(pfile_t *pf)
{
  pic_bank_info_t *bank;
  unsigned         total_avail;
  unsigned         total_used;
  unsigned         ct;

  for (ct = 0, total_used = 0, total_avail = 0, bank = pic_blist;
       bank;
       ct++, total_used += bank->next - bank->lo, 
         total_avail += bank->hi + 1 - bank->lo, bank = bank->link)
    ;
  pfile_log(pf, pfile_log_info, PIC_MSG_DATA_USED,
      total_used, total_avail);
}

void pic_blist_free(pfile_t *pf)
{
  while (pic_blist) {
    pic_bank_info_t *bank;

    bank = pic_blist->link;
    FREE(pic_blist);
    pic_blist = bank;
  }
}

/*
 * NAME
 *   pic_var_ofs_get
 *
 * DESCRIPTION
 *   given a value, determine the true offset
 *
 * PARAMETERS
 *   val : value
 *   ofs : additional offset
 *
 * RETURN
 *   the true offset
 *
 * NOTES
 *   this is for arrays that span banks
 */
variable_base_t pic_var_ofs_get(pfile_t *pf, const value_t *val, size_t ofs)
{
  const pic_bank_info_t *bank;
  const variable_t      *var;
  size_t                 whole_ofs;

  var = value_variable_get(val);

  ofs += value_baseofs_get(val);
  ofs += value_sz_get(val) * value_ct_ofs_get(val);
  whole_ofs = (ofs / variable_sz_get(var)) * variable_sz_get(var);
  ofs -= whole_ofs;
  if (VARIABLE_BASE_UNKNOWN != variable_base_get(var)) {
    size_t var_base; /* lowest point used in the current bank */

    for (bank = pic_blist; bank->hi < variable_base_get(var); bank = bank->link)
      ; /* null body */

    var_base = variable_base_get(var);
    while ((variable_base_get(var) + whole_ofs + variable_sz_get(var)) 
      > (bank->hi+1)) {
      /* we need to add some here */
      size_t var_hi; /* highest element used in this bank */

      var_hi = (bank->hi + 1 - var_base) / variable_sz_get(var);
      var_hi = var_base + var_hi * variable_sz_get(var);
      whole_ofs += (bank->link->lo - var_hi);
      bank = bank->link;
      var_base = bank->lo;
    }
  }
  return whole_ofs + ofs;
}

boolean_t pic_var_spans_banks(pfile_t *pf, const variable_t *var)
{
  boolean_t        ret;

  /* a variable placed at 0 is special -- it directly accesses memory! */
  if ((variable_base_get(var) == VARIABLE_BASE_UNKNOWN) 
    || (0 == variable_base_get(var))) {
    ret = boolean_false;
  } else {
    pic_bank_info_t *bank;
    size_t           sz;

    for (bank = pic_blist; bank->hi < variable_base_get(var); bank = bank->link)
      ; /* null body */

    sz = variable_sz_get(var) * variable_ct_get(var);
    ret = (bank->hi < variable_base_get(var) + sz);
  }
  return ret;
}

size_t pic_var_firstbank_get(pfile_t *pf, const variable_t *var)
{
  size_t n;

  if (!pic_blist) {
    n = 0;
  } else {
    pic_bank_info_t *bank;

    for (bank = pic_blist, n = 0; 
         bank->hi < variable_base_get(var); 
         bank = bank->link, n++)
      ; /* null body */
  }
  return n;
}

/*
 * NAME
 *   pic_var_firstbank_highest_get
 *
 * DESCRIPTION
 *   determine the highest location in the first bank used by a variable
 *
 * PARAMETERS
 *   pf  : 
 *   var : 
 *
 * RETURN
 *
 * NOTES
 *   the highest location used in a bank differs from the real highest
 *   location because an element must fit completely within a bank
 */
size_t pic_var_firstbank_highest_get(pfile_t *pf, const variable_t *var)
{
  pic_bank_info_t *bank;
  variable_base_t  base;
  variable_sz_t    sz;

  base = variable_base_get(var);
  sz   = variable_sz_get(var);

  for (bank = pic_blist; bank->hi < base; bank = bank->link)
    ; /* null body */
  base = base + ((bank->hi + 1 - base) / sz ) * sz;
  if (base + sz - 1 > bank->hi) {
    base--;
  }
  return base;
}

pic_bank_info_t *pic_bank_info_get(pfile_t *pf)
{
  return pic_blist;
}

