/************************************************************
 **
 ** picdelay.c : PIC delay generator
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include "pic.h"
#include "picdelay.h"

/*
 * NAME
 *   pic_delay_create
 *
 * DESCRIPTION
 *   create an n us delay
 *
 * PARAMETERS
 *   pf   :
 *   usec : # of usec to delay
 *
 * RETURN
 *   none
 *
 * NOTES
 *   all delays are inline. There are three used:
 *   1, 2, or 3 variable. all of them force PCLATH<4:3>
 *   to be set always.
 *   Governing equations
 *
 *   1 variable : (3x + 5);        8 instr;  8 <= cycles <=         773
 *   2 variable : y(5x+6) + 3;    14 instr; 14 <= cycles <=     329,216
 *   3 variable : z(y(5x+6)+6)+3; 20 instr; 20 <= cycles <=  84,280,835
 */
static void pic_delay3_create(pfile_t *pf, unsigned long cycles,
    unsigned long *used)
{
  unsigned long x;
  unsigned long best_x;
  unsigned long best_y;
  unsigned long best_z;
  unsigned long best_r;
  label_t      *lblz;
  label_t      *lbly;
  label_t      *lblx;

  best_r = 0;
  best_x = 0;
  best_y = 0;
  best_z = 0;
  for (x = 1; x <= 256; x++) {
    unsigned long y;
    
    for (y = 1; y <= 256; y++) {
      unsigned long z;
      unsigned long r;

      z = (cycles - 3) / ( y * (5 * x + 6) + 6);
      if (z > 256) {
        z = 256;
      }

      r = z * ( y * ( 5 * x + 6) + 6 ) + 3;
      if ((r <= cycles) && (r > best_r)) {
        best_r = r;
        best_x = x;
        best_y = y;
        best_z = z;
      }
    }
  }
  *used = best_z * ( best_y * (5 * best_x + 6) + 6) + 3;
  lblz = 0;
  pfile_label_alloc(pf, 0, 0, &lblz);
  lbly = 0;
  pfile_label_alloc(pf, 0, 0, &lbly);
  lblx = 0;
  pfile_label_alloc(pf, 0, 0, &lblx);

  if (lblz && lbly && lblx) {
    pic_instr_append(pf, pic_opcode_datalo_clr);
    pic_instr_append(pf, pic_opcode_datahi_clr);
    pic_instr_append_w_kn(pf, pic_opcode_movlw, best_z);
    pic_instr_append_reg(pf, pic_opcode_movwf, "_delay2");
    pic_instr_append_label(pf, lblz);
    pic_instr_append_w_kn(pf, pic_opcode_movlw, best_y);
    pic_instr_append_reg(pf, pic_opcode_movwf, "_delay1");
    pic_instr_append_label(pf, lbly);
    pic_instr_append_w_kn(pf, pic_opcode_movlw, best_x);
    pic_instr_append_reg(pf, pic_opcode_movwf, "_delay0");
    pic_instr_append_label(pf, lblx);
    pic_instr_append(pf, pic_opcode_branchhi_clr);
    pic_instr_append(pf, pic_opcode_branchlo_clr);
    pic_instr_append_reg_d(pf, pic_opcode_decfsz, "_delay0", pic_opdst_f);
    pic_instr_append_n(pf, pic_opcode_goto, lblx);
    pic_instr_append(pf, pic_opcode_branchhi_clr);
    pic_instr_append(pf, pic_opcode_branchlo_clr);
    pic_instr_append_reg_d(pf, pic_opcode_decfsz, "_delay1", pic_opdst_f);
    pic_instr_append_n(pf, pic_opcode_goto, lbly);
    pic_instr_append(pf, pic_opcode_branchhi_clr);
    pic_instr_append(pf, pic_opcode_branchlo_clr);
    pic_instr_append_reg_d(pf, pic_opcode_decfsz, "_delay2", pic_opdst_f);
    pic_instr_append_n(pf, pic_opcode_goto, lblz);
  }
  label_release(lblx);
  label_release(lbly);
  label_release(lblz);
}

static void pic_delay2_create(pfile_t *pf, unsigned long cycles,
    unsigned long *used)
{
  unsigned long x;
  unsigned long best_x;
  unsigned long best_y;
  unsigned long best_r;
  label_t      *lbly;
  label_t      *lblx;

  best_r = 0;
  best_x = 0;
  best_y = 0;

  for (x = 1; x <= 256; x++) {
    unsigned long r;
    unsigned long y;

    y = (cycles - 3) / ( 5 * x + 6 );

    if (y > 256) {
      y = 256;
    }
    r = (y * ( 5 * x + 6) + 3);
    if ((r <= cycles) && (r > best_r)) {
      best_r = r;
      best_x = x;
      best_y = y;
    }
  }
  *used = best_y * ( 5 * best_x + 6) + 3;
  /* now that we've a reasonable number, lets use it! */
  lbly = 0;
  pfile_label_alloc(pf, 0, 0, &lbly);
  lblx = 0;
  pfile_label_alloc(pf, 0, 0, &lblx);

  if (lbly && lblx) {
    pic_instr_append(pf, pic_opcode_datalo_clr);
    pic_instr_append(pf, pic_opcode_datahi_clr);
    pic_instr_append_w_kn(pf, pic_opcode_movlw, best_y);
    pic_instr_append_reg(pf, pic_opcode_movwf, "_delay1");
    pic_instr_append_label(pf, lbly);
    pic_instr_append_w_kn(pf, pic_opcode_movlw, best_x);
    pic_instr_append_reg(pf, pic_opcode_movwf, "_delay0");
    pic_instr_append_label(pf, lblx);
    pic_instr_append(pf, pic_opcode_branchhi_clr);
    pic_instr_append(pf, pic_opcode_branchlo_clr);
    pic_instr_append_reg_d(pf, pic_opcode_decfsz, "_delay0", pic_opdst_f);
    pic_instr_append_n(pf, pic_opcode_goto, lblx);
    pic_instr_append(pf, pic_opcode_branchhi_clr);
    pic_instr_append(pf, pic_opcode_branchlo_clr);
    pic_instr_append_reg_d(pf, pic_opcode_decfsz, "_delay1", pic_opdst_f);
    pic_instr_append_n(pf, pic_opcode_goto, lbly);
  }
  label_release(lblx);
  label_release(lbly);
}

void pic_delay_create(pfile_t *pf, variable_const_t usec)
{
  double        cycles;
  unsigned long used;
  value_t      *freq;
  size_t        need;
  variable_t   *var;

  freq = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_freq");
  /* cycles = # of cycles needed */
  cycles = 1.0 * usec * value_const_get(freq) / 4000;
  value_release(freq);
  /*pfile_log(pf, pfile_log_info, "delay %lu requires %.0f cycles", 
      usec, cycles);*/

  /* 329216 is the maximum double-variable delay, so while n
     is greater than this, use the 3-variable delay */
  need = 0;
  while (cycles > 329216UL) {
    need = 3;
    pic_delay3_create(pf, cycles, &used);
    cycles -= used;
    /* pfile_log(pf, pfile_log_info, 
        "...%lu cycles used in 3-var; %.0f remaining", used, cycles);*/
  }
  /* 773 is the maximum single-variable delay, so while n is
     greater than this, use the 2-variable delay */
  while (cycles > 773) {
    if (!need) {
      need = 2;
    }
    pic_delay2_create(pf, cycles, &used);
    cycles -= used;
    /* pfile_log(pf, pfile_log_info, 
        "...%lu cycles used in 2-var; %.0f remaining", used, cycles);*/
  }
  /* 8 is the smallest value worthy of a loop */
  while (cycles >= 8) {
    unsigned long x;
    label_t      *lbl;

    if (!need) {
      need = 1;
    }
    x = (cycles - 5) / 3;
    if (x > 256) {
      x = 256;
    }
    used = (3 * x + 5);
    cycles -= used;
    /* pfile_log(pf, pfile_log_info, 
        "...%lu cycles used in 1-var; %.0f remaining", used, cycles); */
    /* fix this */
    if (result_ok == pfile_label_alloc(pf, 0, 0, &lbl)) {
      pic_instr_append(pf, pic_opcode_datalo_clr);
      pic_instr_append(pf, pic_opcode_datahi_clr);
      pic_instr_append_w_kn(pf, pic_opcode_movlw, (x == 256) ? 0 : x);
      pic_instr_append_reg(pf, pic_opcode_movwf, "_delay0");
      pic_instr_append(pf, pic_opcode_branchhi_clr);
      pic_instr_append(pf, pic_opcode_branchlo_clr);
      pic_instr_append_label(pf, lbl);
      pic_instr_append_reg_d(pf, pic_opcode_decfsz, "_delay0", pic_opdst_f);
      pic_instr_append_n(pf, pic_opcode_goto, lbl);
      label_release(lbl);
    }
  }
  used = cycles;
  while (used--) {
    /* filler */
    pic_instr_append(pf, pic_opcode_nop);
  }
  /* mark the delay variables as needed (or used) */
  switch (need) {
    case 3:
      var = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, 
            "_delay2");
      if (var) {
        variable_sz_set(var, 1);
        variable_release(var);
      }
      /* fall through */
    case 2:
      var = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, 
        "_delay1");
      if (var) {
        variable_sz_set(var, 1);
        variable_release(var);
      }
      /* fall through */
    case 1:
      var = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, 
        "_delay0");
      if (var) {
        variable_sz_set(var, 1);
        variable_release(var);
      }
      /* fall through */
    case 0:
      break;
  }
}

