/************************************************************
 **
 ** slist.c : generic singly linked list definitions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <stdlib.h>
#include <stdio.h>

#include <assert.h>

#include "slist.h"

void slist_init(slist_t *lst)
{
  lst->head = 0;
  lst->tail = 0;
  lst->ct   = 0;
}

void slist_append(slist_t *lst, slist_el_t *el)
{
  slist_insert(lst, lst->tail, el);
}

void slist_insert(slist_t *lst, slist_el_t *after, slist_el_t *el)
{
  slist_el_lock(el);
  if (after) {
    el->link = after->link;
    after->link = el;
    if (!el->link) {
      lst->tail = el;
    }
  } else {
    el->link = lst->head;
    if (!lst->head) {
      lst->tail = el;
    }
    lst->head = el;
  }
  el->owner = lst;
  lst->ct++;
}

void slist_insert_list(slist_t *dst, slist_el_t *after, slist_t *src)
{
  while (src->head) {
    slist_el_t *tmp;

    tmp = src->head;
    slist_el_lock(tmp);
    slist_el_remove(tmp);
    slist_insert(dst, after, tmp);
    after = tmp;
  }
}

void slist_el_remove(slist_el_t *el)
{
  slist_t *lst;

  lst = el->owner;

  if (el == lst->head) {
    lst->head = lst->head->link;
    if (!lst->head) {
      lst->tail = 0;
    }
  } else {
    slist_el_t *elptr;
    
    for (elptr = lst->head; elptr && (elptr->link != el); elptr = elptr->link)
      ;
    if (elptr) {
      elptr->link = el->link;
      if (!elptr->link) {
        lst->tail = elptr;
      }
    }
  }
  el->owner = 0;
  el->link  = 0;
  lst->ct--;
  slist_el_release(el);
}

unsigned slist_ct_get(const slist_t *lst)
{
  return lst->ct;
}


void slist_el_init(slist_el_t *lst, slist_el_free_fn_t fn)
{
  lst->link    = 0;
  lst->owner   = 0;
  lst->ref_ct  = 1;
  lst->free_fn = fn;
}

void slist_el_lock(slist_el_t *el)
{
  ++el->ref_ct;
}

void slist_el_release(slist_el_t *el)
{
  if (el && !--el->ref_ct) {
    assert(0 == el->owner);
    if (el->free_fn) {
      el->free_fn(el);
    }
  }
}

void *slist_el_data(slist_el_t *el)
{
  return (el);
}

void slist_reset(slist_t *lst)
{
  while (lst->head) {
    slist_el_remove(lst->head);
  }
}

slist_el_t *slist_find(slist_t *lst, slist_el_find_fn_t cb, void *arg,
  const void *data)
{
  slist_el_t *el;

  for (el = lst->head; el && cb(arg, el, data); el = el->link)
    ;
  return el;
}

slist_el_t *slist_head(slist_t *lst)
{
  return lst->head;
}

slist_el_t *slist_tail(slist_t *lst)
{
  return lst->tail;
}

slist_el_t *slist_el_next(const slist_el_t *el)
{
  return el->link;
}

refct_t slist_el_refct_get(const slist_el_t *el)
{
  return el->ref_ct;
}

