/**********************************************************
 **
 ** value.c : manipulators for value_t
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include <stdlib.h>
#include <stdio.h>

#include "value.h" 
#include "val_decl.h"

/* notes:
 *   var       -- as expected, the variable
 *   subscript -- for array variables, variable pointing to the element #
 *   baseofs   -- constant subscripts and _h effect base
 *   sz        -- resulting size
 *                operators '_h' and '_l' reduce the size by 1/2
 */
result_t  value_alloc(value_t **dst, variable_t *var, 
   variable_t *subscript, variable_base_t baseofs, variable_sz_t sz)
{
  value_t *val;
  int      rc;

  val = MALLOC(sizeof(*val));
  if (!val) {
    rc = result_memory;
  } else {
    rc = result_ok;
    val->var       = 0;
    val->subscript = 0;
    val->baseofs   = baseofs;
    val->sz        = sz;
    val->ct        = variable_ct_get(var);
    val->ct_ofs    = 0;

    value_variable_set(val, var);
    value_subscript_set(val, subscript);

    val->ref_ct = 1;
    *dst = val;
  }
  return rc;
}

result_t value_clone(value_t **dst, value_t *src)
{
  return value_alloc(dst, src->var, src->subscript, src->baseofs, src->sz);
}


void value_lock(value_t *val)
{
  val->ref_ct++;
}

void value_release(value_t *val)
{
  if (val && !--val->ref_ct) {
    value_subscript_set(val, 0);
    value_variable_set(val, 0);
    FREE(val);
  }
}

void value_sz_set(value_t *val, variable_sz_t sz)
{
  val->sz = sz;
}

variable_sz_t value_sz_get(const value_t *val)
{
  return (val) ? val->sz : 0;
}

boolean_t value_flag_test(const value_t *val, flag_t flag)
{
  return (val) ? variable_flag_test(val->var, flag) : 0;
}

void value_flag_set(value_t *val, flag_t flags)
{
  if (val) {
    variable_flag_set(val->var, flags);
  }
}

void value_flag_clr(value_t *val, flag_t flags)
{
  if (val) {
    variable_flag_clr(val->var, flags);
  }
}

void value_flags_set_all(value_t *val, flag_t flag)
{
  variable_flags_set_all(val->var, flag);
}

flag_t value_flags_get_all(value_t *val)
{
  return (val) ? variable_flags_get_all(val->var) : 0;
}

const char *value_name_get(const value_t *val)
{
  return (val) ? variable_name_get(val->var) : 0;
}

variable_const_t value_const_get(const value_t *val)
{
  variable_const_t n;

  if (!val) {
    n = 0;
  } else {
    variable_ct_t pos;

    pos = val->ct_ofs;
    if (val->subscript) {
      pos += variable_const_get(val->subscript, 0);
    }
    n = variable_const_get(val->var, pos);
    n >>= 8 * val->baseofs;
  }
  return n;
}

void value_const_set(value_t *val, variable_const_t c)
{
  variable_const_set(val->var, 0, c);
}

void value_dump(const value_t *val, FILE *dst)
{
  variable_dump(val->var, dst);
  if (val->subscript || val->baseofs) {
    const char *str;

    fputc('(', dst);
    if (val->subscript && (VALUE_SUBSCRIPT_DUMMY != val->subscript)) {
      variable_dump(val->subscript, dst);
      str = "+";
    } else {
      str = "";
    }
    if (val->baseofs) {
      fprintf(dst, "%s%u", str, val->baseofs);
    }
    fputc(')', dst);
  }
}

void value_expr_ct_bump(value_t *var, ctr_bump_t dir)
{
  variable_expr_ct_bump(var->var, dir);
}

ctr_t value_expr_ct_get(const value_t *val)
{
  return (val) ? variable_expr_ct_get(val->var) : 0;
}

/* two values are the same iff they point to the same variable
 * and have the same offset and subscript */
boolean_t value_same(const value_t *val1, const value_t *val2)
{
  boolean_t rc;

  if ((val1 && !val2) 
    || (!val1 && val2)) {
    rc = boolean_false;
  } else {
    rc = (!val1 && !val2)
      || ((val1->var == val2->var)
        && (val1->subscript == val2->subscript)
        && (val1->baseofs == val2->baseofs));
  }
  return rc;
}

variable_t *value_subscript_get(value_t *val)
{
  return val->subscript;
}

void value_subscript_set(value_t *val, variable_t *subscript)
{
  if (subscript && (VALUE_SUBSCRIPT_DUMMY != subscript)) {
    variable_lock(subscript);
  }
  if (val->subscript && (VALUE_SUBSCRIPT_DUMMY != val->subscript)) {
    variable_release(val->subscript);
  }
  val->subscript = subscript;
}

variable_base_t value_baseofs_get(const value_t *val)
{
  return val->baseofs;
}

void value_baseofs_set(value_t *val, variable_base_t baseofs)
{
  val->baseofs = baseofs;
}

variable_ct_t value_ct_get(const value_t *val)
{
  return (val) ? val->ct : 0;
}

variable_ct_t value_ct_ofs_get(const value_t *val)
{
  return (val) ? val->ct_ofs : 0;
}

void value_ct_ofs_set(value_t *val, variable_ct_t ofs)
{
  val->ct_ofs = ofs;
}

void value_ct_set(value_t *val, variable_ct_t ct)
{
  val->ct = ct;
}


void value_variable_set(value_t *val, variable_t *var)
{
  if (var) {
    variable_lock(var);
  }
  if (val->var) {
    variable_release(val->var);
  }
  val->var = var;
}

variable_t   *value_variable_get(const value_t *val)
{
  return val->var;
}

ctr_t value_assign_ct_get(const value_t *var)
{
  return variable_assign_ct_get(var->var);
}

void value_assign_ct_bump(value_t *var, ctr_bump_t dir)
{
  variable_assign_ct_bump(var->var, dir);
}

ctr_t value_use_ct_get(const value_t *var)
{
  return variable_use_ct_get(var->var);
}

void  value_use_ct_bump(value_t *var, ctr_bump_t dir)
{
  variable_use_ct_bump(var->var, dir);
  if (var->subscript && (VALUE_SUBSCRIPT_DUMMY != var->subscript)) {
    variable_use_ct_bump(var->subscript, dir);
  }
}

boolean_t value_is_const(const value_t *val)
{
  return val
    && (variable_flag_test(val->var, VARIABLE_FLAG_CONST))
    && (!val->subscript 
      || variable_flag_test(val->subscript, VARIABLE_FLAG_CONST));
}

boolean_t value_is_lookup(const value_t *val)
{
  return val
  && (variable_flag_test(val->var, VARIABLE_FLAG_CONST))
  && val->subscript
  && !variable_flag_test(val->subscript,  VARIABLE_FLAG_CONST);
}

boolean_t value_is_array(const value_t *val)
{
  return val && (value_ct_get(val) > 1);
}

variable_base_t value_base_get(const value_t *val)
{
  return (val) ? variable_base_get(val->var) : VARIABLE_BASE_UNKNOWN;
}


